# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for merging and information about it."""

from pathlib import Path

import pytest

from pygit2 import GIT_MERGE_ANALYSIS_UP_TO_DATE
from pygit2 import GIT_MERGE_ANALYSIS_FASTFORWARD
import pygit2


def test_merge_none(mergerepo):
    with pytest.raises(TypeError): mergerepo.merge(None)

def test_merge_analysis_uptodate(mergerepo):
    branch_head_hex = '5ebeeebb320790caf276b9fc8b24546d63316533'
    branch_id = mergerepo.get(branch_head_hex).id

    analysis, preference = mergerepo.merge_analysis(branch_id)
    assert analysis & GIT_MERGE_ANALYSIS_UP_TO_DATE
    assert not analysis & GIT_MERGE_ANALYSIS_FASTFORWARD
    assert {} == mergerepo.status()

    analysis, preference = mergerepo.merge_analysis(branch_id, 'refs/heads/ff-branch')
    assert analysis & GIT_MERGE_ANALYSIS_UP_TO_DATE
    assert not analysis & GIT_MERGE_ANALYSIS_FASTFORWARD
    assert {} == mergerepo.status()

def test_merge_analysis_fastforward(mergerepo):
    branch_head_hex = 'e97b4cfd5db0fb4ebabf4f203979ca4e5d1c7c87'
    branch_id = mergerepo.get(branch_head_hex).id

    analysis, preference = mergerepo.merge_analysis(branch_id)
    assert not analysis & GIT_MERGE_ANALYSIS_UP_TO_DATE
    assert analysis & GIT_MERGE_ANALYSIS_FASTFORWARD
    assert {} == mergerepo.status()

    analysis, preference = mergerepo.merge_analysis(branch_id, 'refs/heads/master')
    assert not analysis & GIT_MERGE_ANALYSIS_UP_TO_DATE
    assert analysis & GIT_MERGE_ANALYSIS_FASTFORWARD
    assert {} == mergerepo.status()

def test_merge_no_fastforward_no_conflicts(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id
    analysis, preference = mergerepo.merge_analysis(branch_id)
    assert not analysis & GIT_MERGE_ANALYSIS_UP_TO_DATE
    assert not analysis & GIT_MERGE_ANALYSIS_FASTFORWARD
    # Asking twice to assure the reference counting is correct
    assert {} == mergerepo.status()
    assert {} == mergerepo.status()

def test_merge_invalid_hex(mergerepo):
    branch_head_hex = '12345678'
    with pytest.raises(KeyError): mergerepo.merge(branch_head_hex)

def test_merge_already_something_in_index(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_oid = mergerepo.get(branch_head_hex).id
    with (Path(mergerepo.workdir) / 'inindex.txt').open('w') as f:
        f.write('new content')
    mergerepo.index.add('inindex.txt')
    with pytest.raises(pygit2.GitError): mergerepo.merge(branch_oid)


def test_merge_no_fastforward_conflicts(mergerepo):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    branch_id = mergerepo.get(branch_head_hex).id

    analysis, preference = mergerepo.merge_analysis(branch_id)
    assert not analysis & GIT_MERGE_ANALYSIS_UP_TO_DATE
    assert not analysis & GIT_MERGE_ANALYSIS_FASTFORWARD

    mergerepo.merge(branch_id)
    assert mergerepo.index.conflicts is not None
    with pytest.raises(KeyError):
        mergerepo.index.conflicts.__getitem__('some-file')

    status = pygit2.GIT_STATUS_CONFLICTED
    # Asking twice to assure the reference counting is correct
    assert {'.gitignore': status} == mergerepo.status()
    assert {'.gitignore': status} == mergerepo.status()

    ancestor, ours, theirs = mergerepo.index.conflicts['.gitignore']
    assert ancestor is None
    assert ours is not None
    assert theirs is not None
    assert '.gitignore' == ours.path
    assert '.gitignore' == theirs.path
    assert 1 == len(list(mergerepo.index.conflicts))

    # Checking the index works as expected
    mergerepo.index.add('.gitignore')
    mergerepo.index.write()
    assert mergerepo.index.conflicts is None
    assert {'.gitignore': pygit2.GIT_STATUS_INDEX_MODIFIED} == mergerepo.status()

def test_merge_remove_conflicts(mergerepo):
    other_branch_tip = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    mergerepo.merge(other_branch_tip)
    idx = mergerepo.index
    conflicts = idx.conflicts
    assert conflicts is not None
    try:
        conflicts['.gitignore']
    except KeyError:
        mergerepo.fail("conflicts['.gitignore'] raised KeyError unexpectedly")
    del idx.conflicts['.gitignore']
    with pytest.raises(KeyError): conflicts.__getitem__('.gitignore')
    assert idx.conflicts is None


def test_merge_commits(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id

    merge_index = mergerepo.merge_commits(mergerepo.head.target, branch_head_hex)
    assert merge_index.conflicts is None
    merge_commits_tree = merge_index.write_tree(mergerepo)

    mergerepo.merge(branch_id)
    index = mergerepo.index
    assert index.conflicts is None
    merge_tree = index.write_tree()

    assert merge_tree == merge_commits_tree

def test_merge_commits_favor(mergerepo):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    merge_index = mergerepo.merge_commits(mergerepo.head.target, branch_head_hex, favor='ours')
    assert merge_index.conflicts is None

    with pytest.raises(ValueError):
        mergerepo.merge_commits(mergerepo.head.target, branch_head_hex, favor='foo')


def test_merge_trees(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id
    ancestor_id = mergerepo.merge_base(mergerepo.head.target, branch_id)

    merge_index = mergerepo.merge_trees(ancestor_id, mergerepo.head.target, branch_head_hex)
    assert merge_index.conflicts is None
    merge_commits_tree = merge_index.write_tree(mergerepo)

    mergerepo.merge(branch_id)
    index = mergerepo.index
    assert index.conflicts is None
    merge_tree = index.write_tree()

    assert merge_tree == merge_commits_tree

def test_merge_trees_favor(mergerepo):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    ancestor_id = mergerepo.merge_base(mergerepo.head.target, branch_head_hex)
    merge_index = mergerepo.merge_trees(ancestor_id, mergerepo.head.target, branch_head_hex, favor='ours')
    assert merge_index.conflicts is None

    with pytest.raises(ValueError):
        mergerepo.merge_trees(ancestor_id, mergerepo.head.target, branch_head_hex, favor='foo')


def test_merge_options():
    from pygit2.ffi import C

    # Default
    o = pygit2.Repository._merge_options()
    assert o.file_favor == C.GIT_MERGE_FILE_FAVOR_NORMAL
    assert o.flags == C.GIT_MERGE_FIND_RENAMES
    assert o.file_flags == 0

    o = pygit2.Repository._merge_options(
        favor='ours', flags={'fail_on_conflict': True}, file_flags={'ignore_whitespace': True}
    )
    assert o.file_favor == C.GIT_MERGE_FILE_FAVOR_OURS
    assert o.flags == C.GIT_MERGE_FIND_RENAMES | C.GIT_MERGE_FAIL_ON_CONFLICT
    assert o.file_flags == C.GIT_MERGE_FILE_IGNORE_WHITESPACE

    o = pygit2.Repository._merge_options(
        favor='theirs', flags={'find_renames': False}, file_flags={'ignore_whitespace': False}
    )
    assert o.file_favor == C.GIT_MERGE_FILE_FAVOR_THEIRS
    assert o.flags == 0
    assert o.file_flags == 0

    o = pygit2.Repository._merge_options(
        favor='union',
        flags={'find_renames': True, 'no_recursive': True},
        file_flags={'diff3_style': True, 'ignore_whitespace': True, 'patience': True}
    )
    assert o.file_favor == C.GIT_MERGE_FILE_FAVOR_UNION
    assert o.flags == C.GIT_MERGE_FIND_RENAMES | C.GIT_MERGE_NO_RECURSIVE
    assert o.file_flags == (
        C.GIT_MERGE_FILE_STYLE_DIFF3
        | C.GIT_MERGE_FILE_IGNORE_WHITESPACE
        | C.GIT_MERGE_FILE_DIFF_PATIENCE
    )


def test_merge_many(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id
    ancestor_id = mergerepo.merge_base_many([mergerepo.head.target, branch_id])

    merge_index = mergerepo.merge_trees(ancestor_id, mergerepo.head.target, branch_head_hex)
    assert merge_index.conflicts is None
    merge_commits_tree = merge_index.write_tree(mergerepo)

    mergerepo.merge(branch_id)
    index = mergerepo.index
    assert index.conflicts is None
    merge_tree = index.write_tree()

    assert merge_tree == merge_commits_tree


def test_merge_octopus(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id
    ancestor_id = mergerepo.merge_base_octopus([mergerepo.head.target, branch_id])

    merge_index = mergerepo.merge_trees(ancestor_id, mergerepo.head.target, branch_head_hex)
    assert merge_index.conflicts is None
    merge_commits_tree = merge_index.write_tree(mergerepo)

    mergerepo.merge(branch_id)
    index = mergerepo.index
    assert index.conflicts is None
    merge_tree = index.write_tree()

    assert merge_tree == merge_commits_tree



from typing import Iterator, Optional
from io import IOBase
from . import Index, Submodule

GIT_APPLY_LOCATION_BOTH: int
GIT_APPLY_LOCATION_INDEX: int
GIT_APPLY_LOCATION_WORKDIR: int
GIT_BLAME_FIRST_PARENT: int
GIT_BLAME_IGNORE_WHITESPACE: int
GIT_BLAME_NORMAL: int
GIT_BLAME_TRACK_COPIES_ANY_COMMIT_COPIES: int
GIT_BLAME_TRACK_COPIES_SAME_COMMIT_COPIES: int
GIT_BLAME_TRACK_COPIES_SAME_COMMIT_MOVES: int
GIT_BLAME_TRACK_COPIES_SAME_FILE: int
GIT_BLAME_USE_MAILMAP: int
GIT_BRANCH_ALL: int
GIT_BRANCH_LOCAL: int
GIT_BRANCH_REMOTE: int
GIT_CHECKOUT_ALLOW_CONFLICTS: int
GIT_CHECKOUT_DISABLE_PATHSPEC_MATCH: int
GIT_CHECKOUT_DONT_UPDATE_INDEX: int
GIT_CHECKOUT_FORCE: int
GIT_CHECKOUT_NONE: int
GIT_CHECKOUT_NO_REFRESH: int
GIT_CHECKOUT_RECREATE_MISSING: int
GIT_CHECKOUT_REMOVE_IGNORED: int
GIT_CHECKOUT_REMOVE_UNTRACKED: int
GIT_CHECKOUT_SAFE: int
GIT_CHECKOUT_SKIP_LOCKED_DIRECTORIES: int
GIT_CHECKOUT_UPDATE_ONLY: int
GIT_CONFIG_LEVEL_GLOBAL: int
GIT_CONFIG_LEVEL_LOCAL: int
GIT_CONFIG_LEVEL_SYSTEM: int
GIT_CONFIG_LEVEL_XDG: int
GIT_DELTA_ADDED: int
GIT_DELTA_CONFLICTED: int
GIT_DELTA_COPIED: int
GIT_DELTA_DELETED: int
GIT_DELTA_IGNORED: int
GIT_DELTA_MODIFIED: int
GIT_DELTA_RENAMED: int
GIT_DELTA_TYPECHANGE: int
GIT_DELTA_UNMODIFIED: int
GIT_DELTA_UNREADABLE: int
GIT_DELTA_UNTRACKED: int
GIT_DESCRIBE_ALL: int
GIT_DESCRIBE_DEFAULT: int
GIT_DESCRIBE_TAGS: int
GIT_DIFF_BREAK_REWRITES: int
GIT_DIFF_BREAK_REWRITES_FOR_RENAMES_ONLY: int
GIT_DIFF_DISABLE_PATHSPEC_MATCH: int
GIT_DIFF_ENABLE_FAST_UNTRACKED_DIRS: int
GIT_DIFF_FIND_ALL: int
GIT_DIFF_FIND_AND_BREAK_REWRITES: int
GIT_DIFF_FIND_BY_CONFIG: int
GIT_DIFF_FIND_COPIES: int
GIT_DIFF_FIND_COPIES_FROM_UNMODIFIED: int
GIT_DIFF_FIND_DONT_IGNORE_WHITESPACE: int
GIT_DIFF_FIND_EXACT_MATCH_ONLY: int
GIT_DIFF_FIND_FOR_UNTRACKED: int
GIT_DIFF_FIND_IGNORE_LEADING_WHITESPACE: int
GIT_DIFF_FIND_IGNORE_WHITESPACE: int
GIT_DIFF_FIND_REMOVE_UNMODIFIED: int
GIT_DIFF_FIND_RENAMES: int
GIT_DIFF_FIND_RENAMES_FROM_REWRITES: int
GIT_DIFF_FIND_REWRITES: int
GIT_DIFF_FLAG_BINARY: int
GIT_DIFF_FLAG_EXISTS: int
GIT_DIFF_FLAG_NOT_BINARY: int
GIT_DIFF_FLAG_VALID_ID: int
GIT_DIFF_FORCE_BINARY: int
GIT_DIFF_FORCE_TEXT: int
GIT_DIFF_IGNORE_CASE: int
GIT_DIFF_IGNORE_FILEMODE: int
GIT_DIFF_IGNORE_SUBMODULES: int
GIT_DIFF_IGNORE_WHITESPACE: int
GIT_DIFF_IGNORE_WHITESPACE_CHANGE: int
GIT_DIFF_IGNORE_WHITESPACE_EOL: int
GIT_DIFF_INCLUDE_CASECHANGE: int
GIT_DIFF_INCLUDE_IGNORED: int
GIT_DIFF_INCLUDE_TYPECHANGE: int
GIT_DIFF_INCLUDE_TYPECHANGE_TREES: int
GIT_DIFF_INCLUDE_UNMODIFIED: int
GIT_DIFF_INCLUDE_UNREADABLE: int
GIT_DIFF_INCLUDE_UNREADABLE_AS_UNTRACKED: int
GIT_DIFF_INCLUDE_UNTRACKED: int
GIT_DIFF_INDENT_HEURISTIC: int
GIT_DIFF_MINIMAL: int
GIT_DIFF_NORMAL: int
GIT_DIFF_PATIENCE: int
GIT_DIFF_RECURSE_IGNORED_DIRS: int
GIT_DIFF_RECURSE_UNTRACKED_DIRS: int
GIT_DIFF_REVERSE: int
GIT_DIFF_SHOW_BINARY: int
GIT_DIFF_SHOW_UNMODIFIED: int
GIT_DIFF_SHOW_UNTRACKED_CONTENT: int
GIT_DIFF_SKIP_BINARY_CHECK: int
GIT_DIFF_STATS_FULL: int
GIT_DIFF_STATS_INCLUDE_SUMMARY: int
GIT_DIFF_STATS_NONE: int
GIT_DIFF_STATS_NUMBER: int
GIT_DIFF_STATS_SHORT: int
GIT_DIFF_UPDATE_INDEX: int
GIT_FILEMODE_BLOB: int
GIT_FILEMODE_BLOB_EXECUTABLE: int
GIT_FILEMODE_COMMIT: int
GIT_FILEMODE_LINK: int
GIT_FILEMODE_TREE: int
GIT_MERGE_ANALYSIS_FASTFORWARD: int
GIT_MERGE_ANALYSIS_NONE: int
GIT_MERGE_ANALYSIS_NORMAL: int
GIT_MERGE_ANALYSIS_UNBORN: int
GIT_MERGE_ANALYSIS_UP_TO_DATE: int
GIT_MERGE_PREFERENCE_FASTFORWARD_ONLY: int
GIT_MERGE_PREFERENCE_NONE: int
GIT_MERGE_PREFERENCE_NO_FASTFORWARD: int
GIT_OBJ_ANY: int
GIT_OBJ_BLOB: int
GIT_OBJ_COMMIT: int
GIT_OBJ_TAG: int
GIT_OBJ_TREE: int
GIT_OID_HEXSZ: int
GIT_OID_HEX_ZERO: str
GIT_OID_MINPREFIXLEN: int
GIT_OID_RAWSZ: int
GIT_OPT_DISABLE_PACK_KEEP_FILE_CHECKS: int
GIT_OPT_ENABLE_CACHING: int
GIT_OPT_ENABLE_FSYNC_GITDIR: int
GIT_OPT_ENABLE_OFS_DELTA: int
GIT_OPT_ENABLE_STRICT_HASH_VERIFICATION: int
GIT_OPT_ENABLE_STRICT_OBJECT_CREATION: int
GIT_OPT_ENABLE_STRICT_SYMBOLIC_REF_CREATION: int
GIT_OPT_ENABLE_UNSAVED_INDEX_SAFETY: int
GIT_OPT_GET_CACHED_MEMORY: int
GIT_OPT_GET_MWINDOW_MAPPED_LIMIT: int
GIT_OPT_GET_MWINDOW_SIZE: int
GIT_OPT_GET_PACK_MAX_OBJECTS: int
GIT_OPT_GET_SEARCH_PATH: int
GIT_OPT_GET_TEMPLATE_PATH: int
GIT_OPT_GET_USER_AGENT: int
GIT_OPT_GET_WINDOWS_SHAREMODE: int
GIT_OPT_SET_ALLOCATOR: int
GIT_OPT_SET_CACHE_MAX_SIZE: int
GIT_OPT_SET_CACHE_OBJECT_LIMIT: int
GIT_OPT_SET_MWINDOW_MAPPED_LIMIT: int
GIT_OPT_SET_MWINDOW_SIZE: int
GIT_OPT_SET_PACK_MAX_OBJECTS: int
GIT_OPT_SET_SEARCH_PATH: int
GIT_OPT_SET_SSL_CERT_LOCATIONS: int
GIT_OPT_SET_SSL_CIPHERS: int
GIT_OPT_SET_TEMPLATE_PATH: int
GIT_OPT_SET_USER_AGENT: int
GIT_OPT_SET_WINDOWS_SHAREMODE: int
GIT_REF_INVALID: int
GIT_REF_LISTALL: int
GIT_REF_OID: int
GIT_REF_SYMBOLIC: int
GIT_RESET_HARD: int
GIT_RESET_MIXED: int
GIT_RESET_SOFT: int
GIT_REVPARSE_MERGE_BASE: int
GIT_REVPARSE_RANGE: int
GIT_REVPARSE_SINGLE: int
GIT_SORT_NONE: int
GIT_SORT_REVERSE: int
GIT_SORT_TIME: int
GIT_SORT_TOPOLOGICAL: int
GIT_STASH_APPLY_DEFAULT: int
GIT_STASH_APPLY_REINSTATE_INDEX: int
GIT_STASH_DEFAULT: int
GIT_STASH_INCLUDE_IGNORED: int
GIT_STASH_INCLUDE_UNTRACKED: int
GIT_STASH_KEEP_INDEX: int
GIT_STATUS_CONFLICTED: int
GIT_STATUS_CURRENT: int
GIT_STATUS_IGNORED: int
GIT_STATUS_INDEX_DELETED: int
GIT_STATUS_INDEX_MODIFIED: int
GIT_STATUS_INDEX_NEW: int
GIT_STATUS_INDEX_RENAMED: int
GIT_STATUS_INDEX_TYPECHANGE: int
GIT_STATUS_WT_DELETED: int
GIT_STATUS_WT_MODIFIED: int
GIT_STATUS_WT_NEW: int
GIT_STATUS_WT_RENAMED: int
GIT_STATUS_WT_TYPECHANGE: int
GIT_STATUS_WT_UNREADABLE: int
LIBGIT2_VERSION: str
LIBGIT2_VER_MAJOR: int
LIBGIT2_VER_MINOR: int
LIBGIT2_VER_REVISION: int

class Object:
    _pointer: bytes
    filemode: int
    hex: str
    id: Oid
    name: str | None
    oid: Oid
    raw_name: bytes | None
    short_id: str
    type: int
    type_str: str
    def peel(self, target_type) -> Object: ...
    def read_raw(self) -> bytes: ...
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __hash__(self) -> int: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class Reference:
    name: str
    raw_name: bytes
    raw_shorthand: bytes
    raw_target: Oid | bytes
    shorthand: str
    target: Oid | str
    type: int
    def __init__(self, *args) -> None: ...
    def delete(self) -> None: ...
    def log(self) -> Iterator[RefLogEntry]: ...
    def peel(self, type = None) -> Object: ...
    def rename(self, new_name: str) -> None: ...
    def resolve(self) -> Reference: ...
    def set_target(self, target: _OidArg, message: str = ...) -> None: ...
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class AlreadyExistsError(ValueError): ...

class Blob(Object):
    data: bytes
    is_binary: bool
    size: int
    def diff(self, blob: Blob = ..., flag: int = ..., old_as_path: str = ..., new_as_path: str = ...) -> Patch: ...
    def diff_to_buffer(self, buffer: bytes = None, flag: int = GIT_DIFF_NORMAL, old_as_path: str = ..., buffer_as_path: str = ...) -> Patch: ...

class Branch(Reference):
    branch_name: str
    raw_branch_name: bytes
    remote_name: str
    upstream: Branch
    upstream_name: str
    def delete(self) -> None: ...
    def is_checked_out(self) -> bool: ...
    def is_head(self) -> bool: ...
    def rename(self, name: str, force: bool = False) -> None: ...

class Commit(Object):
    author: Signature
    commit_time: int
    commit_time_offset: int
    committer: Signature
    gpg_signature: tuple[bytes, bytes]
    message: str
    message_encoding: str
    message_trailers: dict[str, str]
    parent_ids: list[Oid]
    parents: list[Commit]
    raw_message: bytes
    tree: Tree
    tree_id: Oid

class Diff:
    deltas: Iterator[DiffDelta]
    patch: str | None
    patchid: Oid
    stats: DiffStats
    def find_similar(self, flags : int = GIT_DIFF_FIND_BY_CONFIG, rename_threshold : int = 50, copy_threshold : int = 50, rename_from_rewrite_threshold: int = 50, break_rewrite_threshold: int = 60, rename_limit: int = 1000) -> None: ...
    def merge(self, diff: Diff) -> None: ...
    @staticmethod
    def from_c(diff, repo) -> Diff: ...
    @staticmethod
    def parse_diff(git_diff: str | bytes) -> Diff: ...
    def __getitem__(self, index: int) -> Patch: ...  # Diff_getitem
    def __iter__(self) -> Iterator[Patch]: ...  # -> DiffIter
    def __len__(self) -> int: ...

class DiffDelta:
    flags: int
    is_binary: bool
    new_file: DiffFile
    nfiles: int
    old_file: DiffFile
    similarity: int
    status: int
    def status_char(self) -> str: ...

class DiffFile:
    flags: int
    id: Oid
    mode: int
    path: str
    raw_path: bytes
    size: int
    @staticmethod
    def from_c(bytes) -> DiffFile: ...

class DiffHunk:
    header: str
    lines: list[DiffLine]
    new_lines: int
    new_start: int
    old_lines: int
    old_start: int

class DiffLine:
    content: str
    content_offset: int
    new_lineno: int
    num_lines: int
    old_lineno: int
    origin: str
    raw_content: bytes

class DiffStats:
    deletions: int
    files_changed: int
    insertions: int
    def format(self, format: int, width: int) -> str: ...

class GitError(Exception): ...

class InvalidSpecError(ValueError): ...

class Mailmap:
    def __init__(self, *args) -> None: ...
    def add_entry(self, real_name: str = ..., real_email: str = ..., replace_name: str = ..., replace_email: str = ...) -> None: ...
    @staticmethod
    def from_buffer(buffer: str | bytes) -> Mailmap: ...
    @staticmethod
    def from_repository(repository: Repository) -> Mailmap: ...
    def resolve(self, name: str, email: str) -> tuple[str,str]: ...
    def resolve_signature(self, sig: Signature) -> Signature: ...

class Note:
    annotated_id: Oid
    id: Oid
    message: str
    def remove(self, author: Signature, committer: Signature, ref: str = "refs/notes/commits") -> None: ...

class Odb:
    backends: Iterator[OdbBackend]
    def __init__(self, *args, **kwargs) -> None: ...
    def add_backend(self, backend: OdbBackend, priority: int) -> None: ...
    def add_disk_alternate(self, path: str) -> None: ...
    def exists(self, oid: _OidArg) -> bool: ...
    def read(self, oid: _OidArg) -> tuple[int, int, bytes]: ...
    def write(self, type: int, data: bytes) -> Oid: ...
    def __contains__(self, other: _OidArg) -> bool: ...
    def __iter__(self) -> Iterator[Oid]: ...  # Odb_as_iter

class OdbBackend:
    def __init__(self, *args, **kwargs) -> None: ...
    def exists(self, oid: _OidArg) -> bool: ...
    def exists_prefix(self, partial_id: _OidArg) -> Oid: ...
    def read(self, oid: _OidArg) -> tuple[int, bytes]: ...
    def read_header(self, oid: _OidArg) -> tuple[int, int]: ...
    def read_prefix(self, oid: _OidArg) -> tuple[int, bytes, Oid]: ...
    def refresh(self) -> None: ...
    def __iter__(self) -> Iterator[Oid]: ...  # OdbBackend_as_iter

class OdbBackendLoose(OdbBackend):
    def __init__(self, *args, **kwargs) -> None: ...

class OdbBackendPack(OdbBackend):
    def __init__(self, *args, **kwargs) -> None: ...

class Oid:
    hex: str
    raw: bytes
    def __init__(self, raw: bytes = ..., hex: str = ...) -> None: ...
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __hash__(self) -> int: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class Patch:
    data: bytes
    delta: DiffDelta
    hunks: list[DiffHunk]
    line_stats: tuple[int, int, int]  # context, additions, deletions
    text: str | None

    @staticmethod
    def create_from(
            old: Blob | bytes | None,
            new: Blob | bytes | None,
            old_as_path: str = ...,
            new_as_path: str = ...,
            flag: int = ...,
            context_lines: int = ...,
            interhunk_lines: int = ...
    ) -> Patch: ...

class RefLogEntry:
    committer: Signature
    message: str
    oid_new: Oid
    oid_old: Oid
    def __init__(self, *args, **kwargs) -> None: ...

class Refdb:
    def __init__(self, *args, **kwargs) -> None: ...
    def compress(self) -> None: ...
    @staticmethod
    def new(repo: Repository) -> Refdb: ...
    @staticmethod
    def open(repo: Repository) -> Refdb: ...
    def set_backend(self, backend: RefdbBackend) -> None: ...

class RefdbBackend:
    def __init__(self, *args, **kwargs) -> None: ...
    def compress(self) -> None: ...
    def delete(self, ref_name: str, old_id: _OidArg, old_target: str) -> None: ...
    def ensure_log(self, ref_name: str) -> bool: ...
    def exists(self, refname: str) -> bool: ...
    def has_log(self, ref_name: str) -> bool: ...
    def lookup(self, refname: str) -> Reference: ...
    def rename(self, old_name: str, new_name: str, force: bool, who: Signature, message: str) -> Reference: ...
    def write(self, ref: Reference, force: bool, who: Signature, message: str, old: _OidArg, old_target: str) -> None: ...

class RefdbFsBackend(RefdbBackend):
    def __init__(self, *args, **kwargs) -> None: ...

class Repository:
    _pointer: bytes
    default_signature: Signature
    head: Reference
    head_is_detached: bool
    head_is_unborn: bool
    is_bare: bool
    is_empty: bool
    is_shallow: bool
    odb: Odb
    path: str
    refdb: Refdb
    workdir: str
    def __init__(self, *args, **kwargs) -> None: ...
    def TreeBuilder(self, src: Tree | _OidArg = ...) -> TreeBuilder: ...
    def _disown(self, *args, **kwargs) -> None: ...
    def _from_c(self, *args, **kwargs) -> None: ...
    def add_worktree(self, name: str, path: str, ref: Reference = ...) -> Worktree: ...
    def applies(self, diff: Diff, location: int = GIT_APPLY_LOCATION_INDEX, raise_error: bool = False) -> bool: ...
    def apply(self, diff: Diff, location: int = GIT_APPLY_LOCATION_WORKDIR) -> None: ...
    def cherrypick(self, id: _OidArg) -> None: ...
    def compress_references(self) -> None: ...
    def create_blob(self, data: bytes) -> Oid: ...
    def create_blob_fromdisk(self, path: str) -> Oid: ...
    def create_blob_fromiobase(self, iobase: IOBase) -> Oid: ...
    def create_blob_fromworkdir(self, path: str) -> Oid: ...
    def create_branch(self, name: str, commit: Commit, force = False) -> Branch: ...
    def create_commit(self, reference_name: str, author: Signature, committer: Signature, message: str, tree: _OidArg, parents: list[_OidArg], encoding: str = ...) -> Oid: ...
    def create_commit_string(self, author: Signature, committer: Signature, message: str, tree: _OidArg, parents: list[_OidArg], encoding: str = ...) -> Oid: ...
    def create_commit_with_signature(self, content: str, signature: str, signature_field: Optional[str] = None) -> Oid: ...
    def create_note(self, message: str, author: Signature, committer: Signature, annotated_id: str, ref: str = "refs/notes/commits", force: bool = False) -> Oid: ...
    def create_reference_direct(self, name: str, target: _OidArg, force: bool, message: str = None) -> Reference: ...
    def create_reference_symbolic(self, name: str, target: str, force: bool, message: str = None) -> Reference: ...
    def create_tag(self, name: str, oid: _OidArg, type: int, tagger: Signature, message: str) -> Oid: ...
    def descendant_of(self, oid1: _OidArg, oid2: _OidArg) -> bool: ...
    def expand_id(self, hex: str) -> Oid: ...
    def free(self) -> None: ...
    def git_object_lookup_prefix(self, oid: _OidArg) -> Object: ...
    def init_submodules(self, submodules: list[Submodule] = None, overwrite = False) -> None: ...
    def list_worktrees(self) -> list[str]: ...
    def listall_branches(self, flag: int = GIT_BRANCH_LOCAL) -> list[str]: ...
    def listall_reference_objects(self) -> list[Reference]: ...
    def listall_references(self) -> list[str]: ...
    def listall_stashes(self) -> list[Stash]: ...
    def listall_submodules(self) -> list[str]: ...
    def lookup_branch(self, branch_name: str, branch_type: int = GIT_BRANCH_LOCAL) -> Branch: ...
    def lookup_note(self, annotated_id: str, ref: str = "refs/notes/commits") -> Note: ...
    def lookup_reference(self, name: str) -> Reference: ...
    def lookup_reference_dwim(self, name: str) -> Reference: ...
    def lookup_worktree(self, name: str) -> Worktree: ...
    def merge(self, id: _OidArg) -> None: ...
    def merge_analysis(self, their_head: _OidArg, our_ref: str = "HEAD") -> tuple[int,int]: ...
    def merge_base(self, oid1: _OidArg, oid2: _OidArg) -> Oid: ...
    def merge_base_many(self, oids: list[_OidArg]) -> Oid: ...
    def merge_base_octopus(self, oids: list[_OidArg]) -> Oid: ...
    def notes(self) -> Iterator[Note]: ...
    def path_is_ignored(self, path: str) -> bool: ...
    def raw_listall_branches(self, flag: int = GIT_BRANCH_LOCAL) -> list[bytes]: ...
    def raw_listall_references(self) -> list[bytes]: ...
    def reset(self, oid: _OidArg, reset_type: int) -> None: ...
    def revparse(self, revspec: str) -> RevSpec: ...
    def revparse_ext(self, revision: str) -> tuple[Object,Reference]: ...
    def revparse_single(self, revision: str) -> Object: ...
    def set_odb(self, odb: Odb) -> None: ...
    def set_refdb(self, refdb: Refdb) -> None: ...
    def status(self, untracked_files: str = "all", ignored: bool = True) -> dict[str,int]: ...
    def status_file(self, path: str) -> int: ...
    def walk(self, oid: _OidArg | None, sort_mode: int = GIT_SORT_NONE) -> Walker: ...

class RevSpec:
    flags: int
    from_object: Object
    to_object: Object

class Signature:
    _encoding: str | None
    _pointer: bytes
    email: str
    name: str
    offset: int
    raw_email: bytes
    raw_name: bytes
    time: int
    def __init__(self, name: str, email: str, time: int, offset: int, encoding: str | None) -> None: ...
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class Stash:
    commit_id: Oid
    message: str
    raw_message: bytes
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class Tag(Object):
    message: str
    name: str
    raw_message: bytes
    raw_name: bytes
    tagger: Signature
    target: Oid
    def get_object(self) -> object: ...

class Tree(Object):
    def diff_to_index(self, index: Index, flags: int = GIT_DIFF_NORMAL, context_lines: int = 3, interhunk_lines: int = 0) -> Diff: ...
    def diff_to_tree(self, tree: Tree = ..., flags: int = ..., context_lines: int = ..., interhunk_lines: int = ..., swap: bool = ...) -> Diff: ...
    def diff_to_workdir(self, flags: int = GIT_DIFF_NORMAL, context_lines: int = 3, interhunk_lines: int = 0) -> Diff: ...
    def __contains__(self, other: str) -> bool: ...  # Tree_contains
    def __getitem__(self, index: str | int) -> None: ...  # Tree_subscript
    def __iter__(self) -> Iterator[Object]: ...
    def __len__(self) -> int: ...  # Tree_len
    def __rtruediv__(self, other: str) -> Object: ...
    def __truediv__(self, other: str) -> Object: ...  # Tree_divide

class TreeBuilder:
    def clear(self) -> None: ...
    def get(self, name: str) -> Object: ...
    def insert(self, name: str, oid: _OidArg, attr: int) -> None: ...
    def remove(self, name: str) -> None: ...
    def write(self) -> Oid: ...
    def __len__(self) -> int: ...

class Walker:
    def hide(self, oid: _OidArg) -> None: ...
    def push(self, oid: _OidArg) -> None: ...
    def reset(self) -> None: ...
    def simplify_first_parent(self) -> None: ...
    def sort(self, mode: int) -> None: ...
    def __iter__(self) -> Iterator[Commit]: ... #Walker: ...
    def __next__(self) -> Commit: ...

class Worktree:
    is_prunable: bool
    name: str
    path: str
    def prune(self, force = False) -> None: ...

def discover_repository(path: str, across_fs: bool = False, ceiling_dirs: str = ...) -> str: ...
def hash(data: bytes) -> Oid: ...
def hashfile(path: str) -> Oid: ...
def init_file_backend(path: str, flags: int = 0) -> object: ...
def option(*args, **kwargs) -> None: ...
def reference_is_valid_name(refname: str) -> bool: ...
def tree_entry_cmp(a: Object, b: Object) -> int: ...

_OidArg = str | Oid



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

# Standard library
import hashlib
from pathlib import Path
import shutil
import socket
import stat
import sys
import zipfile

# Requirements
import pytest

# Pygit2
import pygit2


requires_future_libgit2 = pytest.mark.skipif(
    pygit2.LIBGIT2_VER < (2, 0, 0),
    reason='This test may work with a future version of libgit2')

try:
    socket.gethostbyname('github.com')
    has_network = True
except socket.gaierror:
    has_network = False

requires_network = pytest.mark.skipif(
    not has_network,
    reason='Requires network')

with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
    has_proxy = sock.connect_ex(('', 8888)) == 0

requires_proxy = pytest.mark.skipif(
    not has_proxy,
    reason='Requires proxy at port 8888')

requires_ssh = pytest.mark.skipif(
    not (pygit2.features & pygit2.GIT_FEATURE_SSH),
    reason='Requires SSH')


is_pypy = '__pypy__' in sys.builtin_module_names

fspath = pytest.mark.skipif(
    is_pypy,
    reason="PyPy doesn't fully support fspath, see https://foss.heptapod.net/pypy/pypy/-/issues/3168")

refcount = pytest.mark.skipif(
    is_pypy,
    reason='skip refcounts checks in pypy')


def force_rm_handle(remove_path, path, excinfo):
    path = Path(path)
    path.chmod(path.stat().st_mode | stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
    remove_path(path)


def gen_blob_sha1(data):
    # http://stackoverflow.com/questions/552659/assigning-git-sha1s-without-git
    m = hashlib.sha1()
    m.update(('blob %d\0' % len(data)).encode())
    m.update(data)
    return m.hexdigest()


def rmtree(path):
    """In Windows a read-only file cannot be removed, and shutil.rmtree fails.
    So we implement our own version of rmtree to address this issue.
    """
    if Path(path).exists():
        onerror = lambda func, path, e: force_rm_handle(func, path, e)
        shutil.rmtree(path, onerror=onerror)


class TemporaryRepository:

    def __init__(self, name, tmp_path):
        self.name = name
        self.tmp_path = tmp_path

    def __enter__(self):
        path = Path(__file__).parent / 'data' / self.name
        temp_repo_path = Path(self.tmp_path) / path.stem
        if path.suffix == '.zip':
            with zipfile.ZipFile(path) as zipf:
                zipf.extractall(self.tmp_path)
        elif path.suffix == '.git':
            shutil.copytree(path, temp_repo_path)
        else:
            raise ValueError(f'Unexpected {path.suffix} extension')

        return temp_repo_path

    def __exit__(self, exc_type, exc_value, traceback):
        pass


def assertRaisesWithArg(exc_class, arg, func, *args, **kwargs):
    with pytest.raises(exc_class) as excinfo:
        func(*args, **kwargs)
    assert excinfo.value.args == (arg,)

    # Explicitly clear the Exception Info. Citing
    # https://docs.pytest.org/en/latest/reference.html#pytest-raises:
    #
    # Clearing those references breaks a reference cycle
    # (ExceptionInfo –> caught exception –> frame stack raising the exception
    # –> current frame stack –> local variables –> ExceptionInfo) which makes
    # Python keep all objects referenced from that cycle (including all local
    # variables in the current frame) alive until the next cyclic garbage
    # collection run. See the official Python try statement documentation for
    # more detailed information.
    del excinfo



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from ._pygit2 import Oid
from .errors import check_error
from .ffi import ffi, C


class Submodule:

    @classmethod
    def _from_c(cls, repo, cptr):
        subm = cls.__new__(cls)

        subm._repo = repo
        subm._subm = cptr

        return subm

    def __del__(self):
        C.git_submodule_free(self._subm)

    def open(self):
        """Open the repository for a submodule."""
        crepo = ffi.new('git_repository **')
        err = C.git_submodule_open(crepo, self._subm)
        check_error(err)

        return self._repo._from_c(crepo[0], True)

    @property
    def name(self):
        """Name of the submodule."""
        name = C.git_submodule_name(self._subm)
        return ffi.string(name).decode('utf-8')

    @property
    def path(self):
        """Path of the submodule."""
        path = C.git_submodule_path(self._subm)
        return ffi.string(path).decode('utf-8')

    @property
    def url(self):
        """URL of the submodule."""
        url = C.git_submodule_url(self._subm)
        return ffi.string(url).decode('utf-8')

    @property
    def branch(self):
        """Branch that is to be tracked by the submodule."""
        branch = C.git_submodule_branch(self._subm)
        return ffi.string(branch).decode('utf-8')

    @property
    def head_id(self):
        """Head of the submodule."""
        head = C.git_submodule_head_id(self._subm)
        return Oid(raw=bytes(ffi.buffer(head)[:]))



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Submodule objects."""

from pathlib import Path

import pygit2
import pytest

from . import utils


SUBM_NAME = 'TestGitRepository'
SUBM_PATH = 'TestGitRepository'
SUBM_URL = 'https://github.com/libgit2/TestGitRepository'
SUBM_HEAD_SHA = '49322bb17d3acc9146f98c97d078513228bbf3c0'


@pytest.fixture
def repo(tmp_path):
    with utils.TemporaryRepository('submodulerepo.zip', tmp_path) as path:
        yield pygit2.Repository(path)


def test_lookup_submodule(repo):
    s = repo.lookup_submodule(SUBM_PATH)
    assert s is not None

def test_lookup_submodule_aspath(repo):
    s = repo.lookup_submodule(Path(SUBM_PATH))
    assert s is not None

def test_listall_submodules(repo):
    submodules = repo.listall_submodules()
    assert len(submodules) == 1
    assert submodules[0] == SUBM_PATH

@utils.requires_network
def test_submodule_open(repo):
    s = repo.lookup_submodule(SUBM_PATH)
    repo.init_submodules()
    repo.update_submodules()
    r = s.open()
    assert r is not None
    assert str(r.head.target) == SUBM_HEAD_SHA

def test_name(repo):
    s = repo.lookup_submodule(SUBM_PATH)
    assert SUBM_NAME == s.name

def test_path(repo):
    s = repo.lookup_submodule(SUBM_PATH)
    assert SUBM_PATH == s.path

def test_url(repo):
    s = repo.lookup_submodule(SUBM_PATH)
    assert SUBM_URL == s.url

@utils.requires_network
def test_init_and_update(repo):
    subrepo_file_path = Path(repo.workdir) / 'TestGitRepository' / 'master.txt'
    assert not subrepo_file_path.exists()
    repo.init_submodules()
    repo.update_submodules()
    assert subrepo_file_path.exists()

@utils.requires_network
def test_specified_update(repo):
    subrepo_file_path = Path(repo.workdir) / 'TestGitRepository' / 'master.txt'
    assert not subrepo_file_path.exists()
    repo.init_submodules(submodules=['TestGitRepository'])
    repo.update_submodules(submodules=['TestGitRepository'])
    assert subrepo_file_path.exists()

@utils.requires_network
def test_oneshot_update(repo):
    subrepo_file_path = Path(repo.workdir) / 'TestGitRepository' / 'master.txt'
    assert not subrepo_file_path.exists()
    repo.update_submodules(init=True)
    assert subrepo_file_path.exists()

@utils.requires_network
def test_head_id(repo):
    s = repo.lookup_submodule(SUBM_PATH)
    assert str(s.head_id) == SUBM_HEAD_SHA

@utils.requires_network
def test_add_submodule(repo):
    sm_repo_path = "test/testrepo"
    sm = repo.add_submodule(SUBM_URL, sm_repo_path)
    sm_repo = sm.open()
    assert sm_repo_path == sm.path
    assert SUBM_URL == sm.url
    assert not sm_repo.is_empty



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import pygit2


expected_diff = b"""diff --git a/iso-8859-1.txt b/iso-8859-1.txt
index e84e339..201e0c9 100644
--- a/iso-8859-1.txt
+++ b/iso-8859-1.txt
@@ -1 +1,2 @@
 Kristian H\xf8gsberg
+foo
"""

def test_patch_from_non_utf8():
    # blobs encoded in ISO-8859-1
    old_content = b'Kristian H\xf8gsberg\n'
    new_content = old_content + b'foo\n'
    patch = pygit2.Patch.create_from(
        old_content,
        new_content,
        old_as_path='iso-8859-1.txt',
        new_as_path='iso-8859-1.txt',
    )

    assert patch.data == expected_diff
    assert patch.text == expected_diff.decode('utf-8', errors='replace')

    # `patch.text` corrupted the ISO-8859-1 content as it forced UTF-8
    # decoding, so assert that we cannot get the original content back:
    assert patch.text.encode('utf-8') != expected_diff

def test_patch_create_from_blobs(encodingrepo):
    patch = pygit2.Patch.create_from(
        encodingrepo['e84e339ac7fcc823106efa65a6972d7a20016c85'],
        encodingrepo['201e0c908e3d9f526659df3e556c3d06384ef0df'],
        old_as_path='iso-8859-1.txt',
        new_as_path='iso-8859-1.txt',
    )

    assert patch.data == expected_diff
    assert patch.text == expected_diff.decode('utf-8', errors='replace')

    # `patch.text` corrupted the ISO-8859-1 content as it forced UTF-8
    # decoding, so assert that we cannot get the original content back:
    assert patch.text.encode('utf-8') != expected_diff



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Remote objects."""

import sys

import pytest

import pygit2
from pygit2 import Oid
from . import utils


REMOTE_NAME = 'origin'
REMOTE_URL = 'https://github.com/libgit2/pygit2.git'
REMOTE_FETCHSPEC_SRC = 'refs/heads/*'
REMOTE_FETCHSPEC_DST = 'refs/remotes/origin/*'
REMOTE_REPO_OBJECTS = 30
REMOTE_REPO_BYTES = 2758

ORIGIN_REFSPEC = '+refs/heads/*:refs/remotes/origin/*'

def test_remote_create(testrepo):
    name = 'upstream'
    url = 'https://github.com/libgit2/pygit2.git'

    remote = testrepo.remotes.create(name, url)

    assert type(remote) == pygit2.Remote
    assert name == remote.name
    assert url == remote.url
    assert remote.push_url is None

    with pytest.raises(ValueError): testrepo.remotes.create(*(name, url))

def test_remote_create_with_refspec(testrepo):
    name = 'upstream'
    url = 'https://github.com/libgit2/pygit2.git'
    fetch = "+refs/*:refs/*"

    remote = testrepo.remotes.create(name, url, fetch)

    assert type(remote) == pygit2.Remote
    assert name == remote.name
    assert url == remote.url
    assert [fetch] == remote.fetch_refspecs
    assert remote.push_url is None

def test_remote_delete(testrepo):
    name = 'upstream'
    url = 'https://github.com/libgit2/pygit2.git'

    testrepo.remotes.create(name, url)
    assert 2 == len(testrepo.remotes)
    remote = testrepo.remotes[1]

    assert name == remote.name
    testrepo.remotes.delete(remote.name)
    assert 1 == len(testrepo.remotes)

def test_remote_rename(testrepo):
    remote = testrepo.remotes[0]

    assert REMOTE_NAME == remote.name
    problems = testrepo.remotes.rename(remote.name, "new")
    assert [] == problems
    assert 'new' != remote.name

    with pytest.raises(ValueError): testrepo.remotes.rename('', '')
    with pytest.raises(ValueError): testrepo.remotes.rename(None, None)


def test_remote_set_url(testrepo):
    remote = testrepo.remotes["origin"]
    assert REMOTE_URL == remote.url

    new_url = 'https://github.com/cholin/pygit2.git'
    testrepo.remotes.set_url("origin", new_url)
    remote = testrepo.remotes["origin"]
    assert new_url == remote.url

    with pytest.raises(ValueError):
        testrepo.remotes.set_url("origin", "")

    testrepo.remotes.set_push_url("origin", new_url)
    remote = testrepo.remotes["origin"]
    assert new_url == remote.push_url
    with pytest.raises(ValueError):
        testrepo.remotes.set_push_url("origin", "")

def test_refspec(testrepo):
    remote = testrepo.remotes["origin"]

    assert remote.refspec_count == 1
    refspec = remote.get_refspec(0)
    assert refspec.src == REMOTE_FETCHSPEC_SRC
    assert refspec.dst == REMOTE_FETCHSPEC_DST
    assert True == refspec.force
    assert ORIGIN_REFSPEC == refspec.string

    assert list == type(remote.fetch_refspecs)
    assert 1 == len(remote.fetch_refspecs)
    assert ORIGIN_REFSPEC == remote.fetch_refspecs[0]

    assert refspec.src_matches('refs/heads/master')
    assert refspec.dst_matches('refs/remotes/origin/master')
    assert 'refs/remotes/origin/master' == refspec.transform('refs/heads/master')
    assert 'refs/heads/master' == refspec.rtransform('refs/remotes/origin/master')

    assert list == type(remote.push_refspecs)
    assert 0 == len(remote.push_refspecs)

    push_specs = remote.push_refspecs
    assert list == type(push_specs)
    assert 0 == len(push_specs)

    testrepo.remotes.add_fetch("origin", '+refs/test/*:refs/test/remotes/*')
    remote = testrepo.remotes["origin"]

    fetch_specs = remote.fetch_refspecs
    assert list == type(fetch_specs)
    assert 2 == len(fetch_specs)
    assert ['+refs/heads/*:refs/remotes/origin/*', '+refs/test/*:refs/test/remotes/*'] == fetch_specs

    testrepo.remotes.add_push("origin", '+refs/test/*:refs/test/remotes/*')

    with pytest.raises(TypeError):
        testrepo.remotes.add_fetch(['+refs/*:refs/*', 5])

    remote = testrepo.remotes["origin"]
    assert ['+refs/test/*:refs/test/remotes/*'] == remote.push_refspecs

def test_remote_list(testrepo):
    assert 1 == len(testrepo.remotes)
    remote = testrepo.remotes[0]
    assert REMOTE_NAME == remote.name
    assert REMOTE_URL == remote.url

    name = 'upstream'
    url = 'https://github.com/libgit2/pygit2.git'
    remote = testrepo.remotes.create(name, url)
    assert remote.name in testrepo.remotes.names()
    assert remote.name in [x.name for x in testrepo.remotes]

@utils.requires_network
def test_ls_remotes(testrepo):
    assert 1 == len(testrepo.remotes)
    remote = testrepo.remotes[0]

    refs = remote.ls_remotes()
    assert refs

    # Check that a known ref is returned.
    assert next(iter(r for r in refs if r['name'] == 'refs/tags/v0.28.2'))

def test_remote_collection(testrepo):
    remote = testrepo.remotes['origin']
    assert REMOTE_NAME == remote.name
    assert REMOTE_URL == remote.url

    with pytest.raises(KeyError):
        testrepo.remotes['upstream']

    name = 'upstream'
    url = 'https://github.com/libgit2/pygit2.git'
    remote = testrepo.remotes.create(name, url)
    assert remote.name in testrepo.remotes.names()
    assert remote.name in [x.name for x in testrepo.remotes]

@utils.refcount
def test_remote_refcount(testrepo):
    start = sys.getrefcount(testrepo)
    remote = testrepo.remotes[0]
    del remote
    end = sys.getrefcount(testrepo)
    assert start == end


def test_fetch(emptyrepo):
    remote = emptyrepo.remotes[0]
    stats = remote.fetch()
    assert stats.received_bytes == REMOTE_REPO_BYTES
    assert stats.indexed_objects == REMOTE_REPO_OBJECTS
    assert stats.received_objects == REMOTE_REPO_OBJECTS

def test_transfer_progress(emptyrepo):
    class MyCallbacks(pygit2.RemoteCallbacks):
        def transfer_progress(emptyrepo, stats):
            emptyrepo.tp = stats

    callbacks = MyCallbacks()
    remote = emptyrepo.remotes[0]
    stats = remote.fetch(callbacks=callbacks)
    assert stats.received_bytes == callbacks.tp.received_bytes
    assert stats.indexed_objects == callbacks.tp.indexed_objects
    assert stats.received_objects == callbacks.tp.received_objects

def test_update_tips(emptyrepo):
    remote = emptyrepo.remotes[0]
    tips = [('refs/remotes/origin/master', Oid(hex='0'*40),
             Oid(hex='784855caf26449a1914d2cf62d12b9374d76ae78')),
            ('refs/tags/root', Oid(hex='0'*40),
             Oid(hex='3d2962987c695a29f1f80b6c3aa4ec046ef44369'))]

    class MyCallbacks(pygit2.RemoteCallbacks):
        def __init__(self, tips):
            self.tips = tips
            self.i = 0

        def update_tips(self, name, old, new):
            assert self.tips[self.i] == (name, old, new)
            self.i += 1

    callbacks = MyCallbacks(tips)
    remote.fetch(callbacks=callbacks)
    assert callbacks.i > 0


@pytest.fixture
def origin(tmp_path):
    with utils.TemporaryRepository('barerepo.zip', tmp_path) as path:
        yield pygit2.Repository(path)

@pytest.fixture
def clone(tmp_path):
    clone = tmp_path / 'clone'
    clone.mkdir()
    with utils.TemporaryRepository('barerepo.zip', clone) as path:
        yield pygit2.Repository(path)

@pytest.fixture
def remote(origin, clone):
    yield clone.remotes.create('origin', origin.path)


def test_push_fast_forward_commits_to_remote_succeeds(origin, clone, remote):
    tip = clone[clone.head.target]
    oid = clone.create_commit(
        'refs/heads/master', tip.author, tip.author, 'empty commit',
        tip.tree.id, [tip.id]
    )
    remote.push(['refs/heads/master'])
    assert origin[origin.head.target].id == oid

def test_push_when_up_to_date_succeeds(origin, clone, remote):
    remote.push(['refs/heads/master'])
    origin_tip = origin[origin.head.target].id
    clone_tip = clone[clone.head.target].id
    assert origin_tip == clone_tip

def test_push_non_fast_forward_commits_to_remote_fails(origin, clone, remote):
    tip = origin[origin.head.target]
    origin.create_commit(
        'refs/heads/master', tip.author, tip.author, 'some commit',
        tip.tree.id, [tip.id]
    )
    tip = clone[clone.head.target]
    clone.create_commit(
        'refs/heads/master', tip.author, tip.author, 'other commit',
        tip.tree.id, [tip.id]
    )

    with pytest.raises(pygit2.GitError):
        remote.push(['refs/heads/master'])



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Object ids."""

# Standard Library
from binascii import unhexlify

from pygit2 import Oid
import pytest


HEX = "15b648aec6ed045b5ca6f57f8b7831a8b4757298"
RAW = unhexlify(HEX.encode('ascii'))


def test_raw():
    oid = Oid(raw=RAW)
    assert oid.raw == RAW
    assert str(oid) == HEX

def test_hex():
    oid = Oid(hex=HEX)
    assert oid.raw == RAW
    assert str(oid) == HEX

def test_hex_bytes():
    hex = bytes(HEX, "ascii")
    with pytest.raises(TypeError): Oid(hex=hex)

def test_none():
    with pytest.raises(ValueError): Oid()

def test_both():
    with pytest.raises(ValueError): Oid(raw=RAW, hex=HEX)

def test_long():
    with pytest.raises(ValueError): Oid(raw=RAW + b'a')
    with pytest.raises(ValueError): Oid(hex=HEX + 'a')

def test_cmp():
    oid1 = Oid(raw=RAW)

    # Equal
    oid2 = Oid(hex=HEX)
    assert oid1 == oid2

    # Not equal
    oid2 = Oid(hex="15b648aec6ed045b5ca6f57f8b7831a8b4757299")
    assert oid1 != oid2

    # Other
    assert oid1 < oid2
    assert oid1 <= oid2
    assert not oid1 == oid2
    assert not oid1 > oid2
    assert not oid1 >= oid2

def test_hash():
    s = set()
    s.add(Oid(raw=RAW))
    s.add(Oid(hex=HEX))
    assert len(s) == 1

    s.add(Oid(hex="0000000000000000000000000000000000000000"))
    s.add(Oid(hex="0000000000000000000000000000000000000001"))
    assert len(s) == 3



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import pygit2
import pytest

from . import utils


@pytest.fixture
def repo(tmp_path):
    with utils.TemporaryRepository('binaryfilerepo.zip', tmp_path) as path:
        yield pygit2.Repository(path)


PATCH_BINARY = """diff --git a/binary_file b/binary_file
index 86e5c10..b835d73 100644
Binary files a/binary_file and b/binary_file differ
"""

PATCH_BINARY_SHOW = """diff --git a/binary_file b/binary_file
index 86e5c1008b5ce635d3e3fffa4434c5eccd8f00b6..b835d73543244b6694f36a8c5dfdffb71b153db7 100644
GIT binary patch
literal 8
Pc${NM%FIhFs^kIy3n&7R

literal 8
Pc${NM&PdElPvrst3ey5{

"""

def test_binary_diff(repo):
    diff = repo.diff('HEAD', 'HEAD^')
    assert PATCH_BINARY == diff.patch
    diff = repo.diff('HEAD', 'HEAD^', flags=pygit2.GIT_DIFF_SHOW_BINARY)
    assert PATCH_BINARY_SHOW == diff.patch
    diff = repo.diff(b'HEAD', b'HEAD^')
    assert PATCH_BINARY == diff.patch
    diff = repo.diff(b'HEAD', b'HEAD^', flags=pygit2.GIT_DIFF_SHOW_BINARY)
    assert PATCH_BINARY_SHOW == diff.patch



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import time

import pytest

from pygit2 import Signature


def __assert(signature, encoding):
    encoding = encoding or 'utf-8'
    assert signature._encoding == encoding
    assert signature.name == signature.raw_name.decode(encoding)
    assert signature.name.encode(encoding) == signature.raw_name
    assert signature.email == signature.raw_email.decode(encoding)
    assert signature.email.encode(encoding) == signature.raw_email

@pytest.mark.parametrize('encoding', [None, 'utf-8', 'iso-8859-1'])
def test_encoding(encoding):
    signature = Signature('Foo Ibáñez', 'foo@example.com', encoding=encoding)
    __assert(signature, encoding)
    assert abs(signature.time - time.time()) < 5
    assert str(signature) == 'Foo Ibáñez <foo@example.com>'

def test_default_encoding():
    signature = Signature('Foo Ibáñez', 'foo@example.com', 1322174594, 60)
    __assert(signature, 'utf-8')

def test_ascii():
    with pytest.raises(UnicodeEncodeError):
        Signature('Foo Ibáñez', 'foo@example.com', encoding='ascii')

def test_repr():
    encoding = 'utf-8'
    signature = Signature('Foo Ibáñez', 'foo@bar.com', 1322174594, 60, encoding=encoding)
    expected = f"pygit2.Signature('Foo Ibáñez', 'foo@bar.com', 1322174594, 60, '{encoding}')"
    assert repr(signature) == expected

def test_repr_from_commit(barerepo):
    repo = barerepo
    signature = Signature('Foo Ibáñez', 'foo@example.com', encoding=None)
    tree = '967fce8df97cc71722d3c2a5930ef3e6f1d27b12'
    parents = ['5fe808e8953c12735680c257f56600cb0de44b10']
    sha = repo.create_commit(None, signature, signature, 'New commit.', tree, parents)
    commit = repo[sha]

    assert repr(signature) == repr(commit.author)
    assert repr(signature) == repr(commit.committer)



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import pygit2
import pytest

import os
from pathlib import Path


def read_content(testrepo):
    with (Path(testrepo.workdir) / 'hello.txt').open('rb') as f:
        return f.read().decode('utf-8')

@pytest.fixture
def new_content():
    content = ['bye world', 'adiós', 'au revoir monde']
    content = ''.join(x + os.linesep for x in content)
    return content

@pytest.fixture
def old_content(testrepo):
    with (Path(testrepo.workdir) / 'hello.txt').open('rb') as f:
        return f.read().decode('utf-8')

@pytest.fixture
def patch_diff(testrepo, new_content):
    # Create the patch
    with (Path(testrepo.workdir) / 'hello.txt').open('wb') as f:
        f.write(new_content.encode('utf-8'))

    patch = testrepo.diff().patch

    # Rollback all changes
    testrepo.checkout('HEAD', strategy=pygit2.GIT_CHECKOUT_FORCE)

    # Return the diff
    return pygit2.Diff.parse_diff(patch)

@pytest.fixture
def foreign_patch_diff():
    patch_contents = """diff --git a/this_file_does_not_exist b/this_file_does_not_exist
index 7f129fd..af431f2 100644
--- a/this_file_does_not_exist
+++ b/this_file_does_not_exist
@@ -1 +1 @@
-a contents 2
+a contents
"""
    return pygit2.Diff.parse_diff(patch_contents)


def test_apply_type_error(testrepo):
    # Check apply type error
    with pytest.raises(TypeError):
        testrepo.apply('HEAD')

def test_apply_diff_to_workdir(testrepo, new_content, patch_diff):
    # Apply the patch and compare
    testrepo.apply(patch_diff, pygit2.GIT_APPLY_LOCATION_WORKDIR)

    assert read_content(testrepo) == new_content
    assert testrepo.status_file('hello.txt') == pygit2.GIT_STATUS_WT_MODIFIED

def test_apply_diff_to_index(testrepo, old_content, patch_diff):
    # Apply the patch and compare
    testrepo.apply(patch_diff, pygit2.GIT_APPLY_LOCATION_INDEX)

    assert read_content(testrepo) == old_content
    assert testrepo.status_file('hello.txt') & pygit2.GIT_STATUS_INDEX_MODIFIED

def test_apply_diff_to_both(testrepo, new_content, patch_diff):
    # Apply the patch and compare
    testrepo.apply(patch_diff, pygit2.GIT_APPLY_LOCATION_BOTH)

    assert read_content(testrepo) == new_content
    assert testrepo.status_file('hello.txt') & pygit2.GIT_STATUS_INDEX_MODIFIED

def test_diff_applies_to_workdir(testrepo, old_content, patch_diff):
    # See if patch applies
    assert testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_WORKDIR)

    # Ensure it was a dry run
    assert read_content(testrepo) == old_content

    # Apply patch for real, then ensure it can't be applied again
    testrepo.apply(patch_diff, pygit2.GIT_APPLY_LOCATION_WORKDIR)
    assert not testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_WORKDIR)

    # It can still be applied to the index, though
    assert testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_INDEX)

def test_diff_applies_to_index(testrepo, old_content, patch_diff):
    # See if patch applies
    assert testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_INDEX)

    # Ensure it was a dry run
    assert read_content(testrepo) == old_content

    # Apply patch for real, then ensure it can't be applied again
    testrepo.apply(patch_diff, pygit2.GIT_APPLY_LOCATION_INDEX)
    assert not testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_INDEX)

    # It can still be applied to the workdir, though
    assert testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_WORKDIR)

def test_diff_applies_to_both(testrepo, old_content, patch_diff):
    # See if patch applies
    assert testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_BOTH)

    # Ensure it was a dry run
    assert read_content(testrepo) == old_content

    # Apply patch for real, then ensure it can't be applied again
    testrepo.apply(patch_diff, pygit2.GIT_APPLY_LOCATION_BOTH)
    assert not testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_BOTH)
    assert not testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_WORKDIR)
    assert not testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_INDEX)

def test_applies_error(testrepo, old_content, patch_diff, foreign_patch_diff):
    # Try to apply a "foreign" patch that affects files that aren't in the repo;
    # ensure we get OSError about the missing file (due to raise_error)
    with pytest.raises(OSError):
        testrepo.applies(foreign_patch_diff, pygit2.GIT_APPLY_LOCATION_BOTH, raise_error=True)

    # Apply a valid patch
    testrepo.apply(patch_diff, pygit2.GIT_APPLY_LOCATION_BOTH)

    # Ensure it can't be applied again and we get an exception about it (due to raise_error)
    with pytest.raises(pygit2.GitError):
        testrepo.applies(patch_diff, pygit2.GIT_APPLY_LOCATION_BOTH, raise_error=True)



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for describing commits."""

import pytest

from pygit2 import GIT_DESCRIBE_TAGS, GIT_DESCRIBE_ALL
import pygit2


def add_tag(repo, name, target):
    message = 'Example tag.\n'
    tagger = pygit2.Signature('John Doe', 'jdoe@example.com', 12347, 0)

    sha = repo.create_tag(name, target, pygit2.GIT_OBJ_COMMIT, tagger, message)
    return sha


def test_describe(testrepo):
    add_tag(testrepo, 'thetag', '4ec4389a8068641da2d6578db0419484972284c8')
    assert 'thetag-2-g2be5719' == testrepo.describe()

def test_describe_without_ref(testrepo):
    with pytest.raises(pygit2.GitError): testrepo.describe()

def test_describe_default_oid(testrepo):
    assert '2be5719' == testrepo.describe(show_commit_oid_as_fallback=True)

def test_describe_strategies(testrepo):
    assert 'heads/master' == testrepo.describe(describe_strategy=GIT_DESCRIBE_ALL)

    testrepo.create_reference('refs/tags/thetag', '4ec4389a8068641da2d6578db0419484972284c8')
    with pytest.raises(KeyError): testrepo.describe()
    assert 'thetag-2-g2be5719' == testrepo.describe(describe_strategy=GIT_DESCRIBE_TAGS)

def test_describe_pattern(testrepo):
    add_tag(testrepo, 'private/tag1', '5ebeeebb320790caf276b9fc8b24546d63316533')
    add_tag(testrepo, 'public/tag2', '4ec4389a8068641da2d6578db0419484972284c8')

    assert 'public/tag2-2-g2be5719' == testrepo.describe(pattern='public/*')

def test_describe_committish(testrepo):
    add_tag(testrepo, 'thetag', 'acecd5ea2924a4b900e7e149496e1f4b57976e51')
    assert 'thetag-4-g2be5719' == testrepo.describe(committish='HEAD')
    assert 'thetag-1-g5ebeeeb' == testrepo.describe(committish='HEAD^')

    assert 'thetag-4-g2be5719' == testrepo.describe(committish=testrepo.head)

    assert 'thetag-1-g6aaa262' == testrepo.describe(committish='6aaa262e655dd54252e5813c8e5acd7780ed097d')
    assert 'thetag-1-g6aaa262' == testrepo.describe(committish='6aaa262')

def test_describe_follows_first_branch_only(testrepo):
    add_tag(testrepo, 'thetag', '4ec4389a8068641da2d6578db0419484972284c8')
    with pytest.raises(KeyError):
        testrepo.describe(only_follow_first_parent=True)

def test_describe_abbreviated_size(testrepo):
    add_tag(testrepo, 'thetag', '4ec4389a8068641da2d6578db0419484972284c8')
    assert 'thetag-2-g2be5719152d4f82c' == testrepo.describe(abbreviated_size=16)
    assert 'thetag' == testrepo.describe(abbreviated_size=0)

def test_describe_long_format(testrepo):
    add_tag(testrepo, 'thetag', '2be5719152d4f82c7302b1c0932d8e5f0a4a0e98')
    assert 'thetag-0-g2be5719' == testrepo.describe(always_use_long_format=True)



def test_describe_dirty(dirtyrepo):
    add_tag(dirtyrepo, 'thetag', 'a763aa560953e7cfb87ccbc2f536d665aa4dff22')
    assert 'thetag' == dirtyrepo.describe()

def test_describe_dirty_with_suffix(dirtyrepo):
    add_tag(dirtyrepo, 'thetag', 'a763aa560953e7cfb87ccbc2f536d665aa4dff22')
    assert 'thetag-dirty' == dirtyrepo.describe(dirty_suffix='-dirty')



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import pygit2
import pytest


@pytest.fixture
def clonerepo(testrepo, tmp_path):
    cloned_repo_path = tmp_path / 'test_remote_prune'

    pygit2.clone_repository(testrepo.workdir, cloned_repo_path)
    clonerepo = pygit2.Repository(cloned_repo_path)
    testrepo.branches.delete('i18n')
    yield clonerepo


def test_fetch_remote_default(clonerepo):
    clonerepo.remotes[0].fetch()
    assert 'origin/i18n' in clonerepo.branches

def test_fetch_remote_prune(clonerepo):
    clonerepo.remotes[0].fetch(prune=pygit2.GIT_FETCH_PRUNE)
    assert 'origin/i18n' not in clonerepo.branches

def test_fetch_no_prune(clonerepo):
    clonerepo.remotes[0].fetch(prune=pygit2.GIT_FETCH_NO_PRUNE)
    assert 'origin/i18n' in clonerepo.branches

def test_remote_prune(clonerepo):
    pruned = []
    class MyCallbacks(pygit2.RemoteCallbacks):
        def update_tips(self, name, old, new):
            pruned.append(name)

    callbacks = MyCallbacks()
    remote = clonerepo.remotes['origin']
    # We do a fetch in order to establish the connection to the remote.
    # Prune operation requires an active connection.
    remote.fetch(prune=pygit2.GIT_FETCH_NO_PRUNE)
    assert 'origin/i18n' in clonerepo.branches
    remote.prune(callbacks)
    assert pruned == ['refs/remotes/origin/i18n']
    assert 'origin/i18n' not in clonerepo.branches



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import pygit2
import pytest


BLOB_OLD_SHA = 'a520c24d85fbfc815d385957eed41406ca5a860b'
BLOB_NEW_SHA = '3b18e512dba79e4c8300dd08aeb37f8e728b8dad'
BLOB_OLD_CONTENT = b"""hello world
hola mundo
bonjour le monde
"""
BLOB_NEW_CONTENT = b'foo bar\n'

BLOB_OLD_PATH = 'a/file'
BLOB_NEW_PATH = 'b/file'

BLOB_PATCH2 = """diff --git a/a/file b/b/file
index a520c24..3b18e51 100644
--- a/a/file
+++ b/b/file
@@ -1,3 +1 @@
 hello world
-hola mundo
-bonjour le monde
"""

BLOB_PATCH = """diff --git a/a/file b/b/file
index a520c24..d675fa4 100644
--- a/a/file
+++ b/b/file
@@ -1,3 +1 @@
-hello world
-hola mundo
-bonjour le monde
+foo bar
"""

BLOB_PATCH_ADDED = """diff --git a/a/file b/b/file
new file mode 100644
index 0000000..d675fa4
--- /dev/null
+++ b/b/file
@@ -0,0 +1 @@
+foo bar
"""

BLOB_PATCH_DELETED = """diff --git a/a/file b/b/file
deleted file mode 100644
index a520c24..0000000
--- a/a/file
+++ /dev/null
@@ -1,3 +0,0 @@
-hello world
-hola mundo
-bonjour le monde
"""


def test_patch_create_from_buffers():
    patch = pygit2.Patch.create_from(
        BLOB_OLD_CONTENT,
        BLOB_NEW_CONTENT,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH

def test_patch_create_from_blobs(testrepo):
    old_blob = testrepo[BLOB_OLD_SHA]
    new_blob = testrepo[BLOB_NEW_SHA]

    patch = pygit2.Patch.create_from(
        old_blob,
        new_blob,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH2

def test_patch_create_from_blob_buffer(testrepo):
    old_blob = testrepo[BLOB_OLD_SHA]
    patch = pygit2.Patch.create_from(
        old_blob,
        BLOB_NEW_CONTENT,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH

def test_patch_create_from_blob_buffer_add(testrepo):
    patch = pygit2.Patch.create_from(
        None,
        BLOB_NEW_CONTENT,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH_ADDED

def test_patch_create_from_blob_buffer_delete(testrepo):
    old_blob = testrepo[BLOB_OLD_SHA]

    patch = pygit2.Patch.create_from(
        old_blob,
        None,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH_DELETED

def test_patch_create_from_bad_old_type_arg(testrepo):
    with pytest.raises(TypeError):
        pygit2.Patch.create_from(testrepo, BLOB_NEW_CONTENT)

def test_patch_create_from_bad_new_type_arg(testrepo):
    with pytest.raises(TypeError):
        pygit2.Patch.create_from(None, testrepo)

def test_context_lines(testrepo):
    old_blob = testrepo[BLOB_OLD_SHA]
    new_blob = testrepo[BLOB_NEW_SHA]

    patch = pygit2.Patch.create_from(
        old_blob,
        new_blob,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    context_count = len(
        [line for line in patch.text.splitlines() if line.startswith(" ")]
    )

    assert context_count != 0

def test_no_context_lines(testrepo):
    old_blob = testrepo[BLOB_OLD_SHA]
    new_blob = testrepo[BLOB_NEW_SHA]

    patch = pygit2.Patch.create_from(
        old_blob,
        new_blob,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
        context_lines=0,
    )

    context_count = len(
        [line for line in patch.text.splitlines() if line.startswith(" ")]
    )

    assert context_count == 0

def test_patch_create_blob_blobs(testrepo):
    old_blob = testrepo[testrepo.create_blob(BLOB_OLD_CONTENT)]
    new_blob = testrepo[testrepo.create_blob(BLOB_NEW_CONTENT)]

    patch = pygit2.Patch.create_from(
        old_blob,
        new_blob,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH

def test_patch_create_blob_buffer(testrepo):
    blob = testrepo[testrepo.create_blob(BLOB_OLD_CONTENT)]
    patch = pygit2.Patch.create_from(
        blob,
        BLOB_NEW_CONTENT,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH

def test_patch_create_blob_delete(testrepo):
    blob = testrepo[testrepo.create_blob(BLOB_OLD_CONTENT)]
    patch = pygit2.Patch.create_from(
        blob,
        None,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH_DELETED

def test_patch_create_blob_add(testrepo):
    blob = testrepo[testrepo.create_blob(BLOB_NEW_CONTENT)]
    patch = pygit2.Patch.create_from(
        None,
        blob,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    assert patch.text == BLOB_PATCH_ADDED

def test_patch_delete_blob(testrepo):
    blob = testrepo[BLOB_OLD_SHA]
    patch = pygit2.Patch.create_from(
        blob,
        None,
        old_as_path=BLOB_OLD_PATH,
        new_as_path=BLOB_NEW_PATH,
    )

    # Make sure that even after deleting the blob the patch still has the
    # necessary references to generate its patch
    del blob
    assert patch.text == BLOB_PATCH_DELETED

def test_patch_multi_blob(testrepo):
    blob = testrepo[BLOB_OLD_SHA]
    patch = pygit2.Patch.create_from(
        blob,
        None
    )
    patch_text = patch.text

    blob = testrepo[BLOB_OLD_SHA]
    patch2 = pygit2.Patch.create_from(
        blob,
        None
    )
    patch_text2 = patch.text

    assert patch_text == patch_text2
    assert patch_text == patch.text
    assert patch_text2 == patch2.text
    assert patch.text == patch2.text



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

# Standard Library
import binascii
import os
from pathlib import Path
import sys
import tempfile

import pygit2
import pytest

from . import utils


HEAD_SHA = '784855caf26449a1914d2cf62d12b9374d76ae78'
PARENT_SHA = 'f5e5aa4e36ab0fe62ee1ccc6eb8f79b866863b87'  # HEAD^
BLOB_HEX = 'af431f20fc541ed6d5afede3e2dc7160f6f01f16'
BLOB_RAW = binascii.unhexlify(BLOB_HEX.encode('ascii'))
BLOB_OID = pygit2.Oid(raw=BLOB_RAW)

def test_is_empty(barerepo):
    assert not barerepo.is_empty

def test_is_bare(barerepo):
    assert barerepo.is_bare

def test_head(barerepo):
    head = barerepo.head
    assert HEAD_SHA == head.target.hex
    assert type(head) == pygit2.Reference
    assert not barerepo.head_is_unborn
    assert not barerepo.head_is_detached

def test_set_head(barerepo):
    # Test setting a detatched HEAD.
    barerepo.set_head(pygit2.Oid(hex=PARENT_SHA))
    assert barerepo.head.target.hex == PARENT_SHA
    # And test setting a normal HEAD.
    barerepo.set_head("refs/heads/master")
    assert barerepo.head.name == "refs/heads/master"
    assert barerepo.head.target.hex == HEAD_SHA

def test_read(barerepo):
    with pytest.raises(TypeError):
        barerepo.read(123)
    utils.assertRaisesWithArg(KeyError, '1' * 40, barerepo.read, '1' * 40)

    ab = barerepo.read(BLOB_OID)
    a = barerepo.read(BLOB_HEX)
    assert ab == a
    assert (pygit2.GIT_OBJ_BLOB, b'a contents\n') == a

    a2 = barerepo.read('7f129fd57e31e935c6d60a0c794efe4e6927664b')
    assert (pygit2.GIT_OBJ_BLOB, b'a contents 2\n') == a2

    a_hex_prefix = BLOB_HEX[:4]
    a3 = barerepo.read(a_hex_prefix)
    assert (pygit2.GIT_OBJ_BLOB, b'a contents\n') == a3

def test_write(barerepo):
    data = b"hello world"
    # invalid object type
    with pytest.raises(ValueError):
        barerepo.write(pygit2.GIT_OBJ_ANY, data)

    oid = barerepo.write(pygit2.GIT_OBJ_BLOB, data)
    assert type(oid) == pygit2.Oid

def test_contains(barerepo):
    with pytest.raises(TypeError):
        123 in barerepo
    assert BLOB_OID in barerepo
    assert BLOB_HEX in barerepo
    assert BLOB_HEX[:10] in barerepo
    assert ('a' * 40) not in barerepo
    assert ('a' * 20) not in barerepo

def test_iterable(barerepo):
    l = [obj for obj in barerepo]
    oid = pygit2.Oid(hex=BLOB_HEX)
    assert oid in l

def test_lookup_blob(barerepo):
    with pytest.raises(TypeError):
        barerepo[123]
    assert barerepo[BLOB_OID].hex == BLOB_HEX
    a = barerepo[BLOB_HEX]
    assert b'a contents\n' == a.read_raw()
    assert BLOB_HEX == a.hex
    assert pygit2.GIT_OBJ_BLOB == a.type

def test_lookup_blob_prefix(barerepo):
    a = barerepo[BLOB_HEX[:5]]
    assert b'a contents\n' == a.read_raw()
    assert BLOB_HEX == a.hex
    assert pygit2.GIT_OBJ_BLOB == a.type

def test_lookup_commit(barerepo):
    commit_sha = '5fe808e8953c12735680c257f56600cb0de44b10'
    commit = barerepo[commit_sha]
    assert commit_sha == commit.hex
    assert pygit2.GIT_OBJ_COMMIT == commit.type
    assert commit.message == ('Second test data commit.\n\n'
                              'This commit has some additional text.\n')

def test_lookup_commit_prefix(barerepo):
    commit_sha = '5fe808e8953c12735680c257f56600cb0de44b10'
    commit_sha_prefix = commit_sha[:7]
    too_short_prefix = commit_sha[:3]
    commit = barerepo[commit_sha_prefix]
    assert commit_sha == commit.hex
    assert pygit2.GIT_OBJ_COMMIT == commit.type
    assert 'Second test data commit.\n\n' 'This commit has some additional text.\n' == commit.message
    with pytest.raises(ValueError):
        barerepo.__getitem__(too_short_prefix)

def test_expand_id(barerepo):
    commit_sha = '5fe808e8953c12735680c257f56600cb0de44b10'
    expanded = barerepo.expand_id(commit_sha[:7])
    assert commit_sha == expanded.hex

@utils.refcount
def test_lookup_commit_refcount(barerepo):
    start = sys.getrefcount(barerepo)
    commit_sha = '5fe808e8953c12735680c257f56600cb0de44b10'
    commit = barerepo[commit_sha]
    del commit
    end = sys.getrefcount(barerepo)
    assert start == end

def test_get_path(barerepo_path):
    barerepo, path = barerepo_path

    directory = Path(barerepo.path).resolve()
    assert directory == path.resolve()

def test_get_workdir(barerepo):
    assert barerepo.workdir is None

def test_revparse_single(barerepo):
    parent = barerepo.revparse_single('HEAD^')
    assert parent.hex == PARENT_SHA

def test_hash(barerepo):
    data = "foobarbaz"
    hashed_sha1 = pygit2.hash(data)
    written_sha1 = barerepo.create_blob(data)
    assert hashed_sha1 == written_sha1

def test_hashfile(barerepo):
    data = "bazbarfoo"
    handle, tempfile_path = tempfile.mkstemp()
    with os.fdopen(handle, 'w') as fh:
        fh.write(data)
    hashed_sha1 = pygit2.hashfile(tempfile_path)
    Path(tempfile_path).unlink()
    written_sha1 = barerepo.create_blob(data)
    assert hashed_sha1 == written_sha1

def test_conflicts_in_bare_repository(barerepo):
    def create_conflict_file(repo, branch, content):
        oid = repo.create_blob(content.encode('utf-8'))
        tb = repo.TreeBuilder()
        tb.insert('conflict', oid, pygit2.GIT_FILEMODE_BLOB)
        tree = tb.write()

        sig = pygit2.Signature('Author', 'author@example.com')
        commit = repo.create_commit(branch.name, sig, sig,
                'Conflict', tree, [branch.target])
        assert commit is not None
        return commit

    b1 = barerepo.create_branch('b1', barerepo.head.peel())
    c1 = create_conflict_file(barerepo, b1, 'ASCII - abc')
    b2 = barerepo.create_branch('b2', barerepo.head.peel())
    c2 = create_conflict_file(barerepo, b2, 'Unicode - äüö')

    index = barerepo.merge_commits(c1, c2)
    assert index.conflicts is not None

    # ConflictCollection does not allow calling len(...) on it directly so
    # we have to calculate length by iterating over its entries
    assert sum(1 for _ in index.conflicts) == 1

    (a, t, o) = index.conflicts['conflict']
    diff = barerepo.merge_file_from_index(a, t, o)
    assert diff == '''<<<<<<< conflict
ASCII - abc
=======
Unicode - äüö
>>>>>>> conflict
'''



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Blame objects."""

import pytest

from pygit2 import Signature, Oid, GIT_BLAME_IGNORE_WHITESPACE


PATH = 'hello.txt'

HUNKS = [
    (Oid(hex='acecd5ea2924a4b900e7e149496e1f4b57976e51'), 1,
     Signature('J. David Ibañez', 'jdavid@itaapy.com',
               1297179898, 60, encoding='utf-8'), True),
    (Oid(hex='6aaa262e655dd54252e5813c8e5acd7780ed097d'), 2,
     Signature('J. David Ibañez', 'jdavid@itaapy.com',
               1297696877, 60, encoding='utf-8'), False),
    (Oid(hex='4ec4389a8068641da2d6578db0419484972284c8'), 3,
     Signature('J. David Ibañez', 'jdavid@itaapy.com',
               1297696908, 60, encoding='utf-8'), False)
]


def test_blame_index(testrepo):
    blame = testrepo.blame(PATH)

    assert len(blame) == 3

    for i, hunk in enumerate(blame):
        assert hunk.lines_in_hunk == 1
        assert HUNKS[i][0] == hunk.final_commit_id
        assert HUNKS[i][1] == hunk.final_start_line_number
        assert HUNKS[i][2] == hunk.final_committer
        assert HUNKS[i][0] == hunk.orig_commit_id
        assert hunk.orig_path == PATH
        assert HUNKS[i][1] == hunk.orig_start_line_number
        assert HUNKS[i][2] == hunk.orig_committer
        assert HUNKS[i][3] == hunk.boundary


def test_blame_flags(blameflagsrepo):
    blame = blameflagsrepo.blame(PATH, flags=GIT_BLAME_IGNORE_WHITESPACE)

    assert len(blame) == 3

    for i, hunk in enumerate(blame):
        assert hunk.lines_in_hunk == 1
        assert HUNKS[i][0] == hunk.final_commit_id
        assert HUNKS[i][1] == hunk.final_start_line_number
        assert HUNKS[i][2] == hunk.final_committer
        assert HUNKS[i][0] == hunk.orig_commit_id
        assert hunk.orig_path == PATH
        assert HUNKS[i][1] == hunk.orig_start_line_number
        assert HUNKS[i][2] == hunk.orig_committer
        assert HUNKS[i][3] == hunk.boundary


def test_blame_with_invalid_index(testrepo):
    blame = testrepo.blame(PATH)

    def test():
        blame[100000]
        blame[-1]

    with pytest.raises(IndexError): test()

def test_blame_for_line(testrepo):
    blame = testrepo.blame(PATH)

    for i, line in zip(range(0, 2), range(1, 3)):
        hunk = blame.for_line(line)

        assert hunk.lines_in_hunk == 1
        assert HUNKS[i][0] == hunk.final_commit_id
        assert HUNKS[i][1] == hunk.final_start_line_number
        assert HUNKS[i][2] == hunk.final_committer
        assert HUNKS[i][0] == hunk.orig_commit_id
        assert hunk.orig_path == PATH
        assert HUNKS[i][1] == hunk.orig_start_line_number
        assert HUNKS[i][2] == hunk.orig_committer
        assert HUNKS[i][3] == hunk.boundary

def test_blame_with_invalid_line(testrepo):
    blame = testrepo.blame(PATH)

    def test():
        blame.for_line(0)
        blame.for_line(100000)
        blame.for_line(-1)

    with pytest.raises(IndexError): test()

def test_blame_newest(testrepo):
    revs = [
        ( 'master^2',   3 ),
        ( 'master^2^',  2 ),
        ( 'master^2^^', 1 ),
    ]

    for rev, num_commits in revs:
        commit = testrepo.revparse_single(rev)
        blame = testrepo.blame(PATH, newest_commit=commit.id)

        assert len(blame) == num_commits

        for i, hunk in enumerate(tuple(blame)[:num_commits]):
            assert hunk.lines_in_hunk == 1
            assert HUNKS[i][0] == hunk.final_commit_id
            assert HUNKS[i][1] == hunk.final_start_line_number
            assert HUNKS[i][2] == hunk.final_committer
            assert HUNKS[i][0] == hunk.orig_commit_id
            assert hunk.orig_path == PATH
            assert HUNKS[i][1] == hunk.orig_start_line_number
            assert HUNKS[i][2] == hunk.orig_committer
            assert HUNKS[i][3] == hunk.boundary



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Tag objects."""

import pygit2
import pytest


TAG_SHA = '3d2962987c695a29f1f80b6c3aa4ec046ef44369'


def test_read_tag(barerepo):
    repo = barerepo
    tag = repo[TAG_SHA]
    target = repo[tag.target]
    assert isinstance(tag, pygit2.Tag)
    assert pygit2.GIT_OBJ_TAG == tag.type
    assert pygit2.GIT_OBJ_COMMIT == target.type
    assert 'root' == tag.name
    assert 'Tagged root commit.\n' == tag.message
    assert 'Initial test data commit.\n' == target.message
    assert tag.tagger == pygit2.Signature('Dave Borowitz', 'dborowitz@google.com', 1288724692, -420)

def test_new_tag(barerepo):
    name = 'thetag'
    target = 'af431f20fc541ed6d5afede3e2dc7160f6f01f16'
    message = 'Tag a blob.\n'
    tagger = pygit2.Signature('John Doe', 'jdoe@example.com', 12347, 0)

    target_prefix = target[:5]
    too_short_prefix = target[:3]
    with pytest.raises(ValueError):
        barerepo.create_tag(name, too_short_prefix, pygit2.GIT_OBJ_BLOB, tagger, message)

    sha = barerepo.create_tag(name, target_prefix, pygit2.GIT_OBJ_BLOB,
                               tagger, message)
    tag = barerepo[sha]

    assert '3ee44658fd11660e828dfc96b9b5c5f38d5b49bb' == tag.hex
    assert name == tag.name
    assert target == tag.target.hex
    assert tagger == tag.tagger
    assert message == tag.message
    assert name == barerepo[tag.hex].name

def test_modify_tag(barerepo):
    name = 'thetag'
    target = 'af431f20fc541ed6d5afede3e2dc7160f6f01f16'
    message = 'Tag a blob.\n'
    tagger = ('John Doe', 'jdoe@example.com', 12347)

    tag = barerepo[TAG_SHA]
    with pytest.raises(AttributeError): setattr(tag, 'name', name)
    with pytest.raises(AttributeError): setattr(tag, 'target', target)
    with pytest.raises(AttributeError): setattr(tag, 'tagger', tagger)
    with pytest.raises(AttributeError): setattr(tag, 'message', message)

def test_get_object(barerepo):
    repo = barerepo
    tag = repo[TAG_SHA]
    assert repo[tag.target].id == tag.get_object().id



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Odb objects."""

# Standard Library
import binascii
from pathlib import Path

import pytest

# pygit2
from pygit2 import Odb, Oid
from pygit2 import GIT_OBJ_ANY, GIT_OBJ_BLOB
from . import utils


BLOB_HEX = 'af431f20fc541ed6d5afede3e2dc7160f6f01f16'
BLOB_RAW = binascii.unhexlify(BLOB_HEX.encode('ascii'))
BLOB_OID = Oid(raw=BLOB_RAW)


def test_emptyodb(barerepo):
    odb = Odb()

    assert len([str(o) for o in odb]) == 0
    assert BLOB_HEX not in odb
    path = Path(barerepo.path) / 'objects'
    odb.add_disk_alternate(path)
    assert BLOB_HEX in odb


@pytest.fixture
def odb(barerepo):
    odb = barerepo.odb
    yield odb

def test_iterable(odb):
    assert BLOB_HEX in [str(o) for o in odb]

def test_contains(odb):
    assert BLOB_HEX in odb

def test_read(odb):
    with pytest.raises(TypeError):
        odb.read(123)
    utils.assertRaisesWithArg(KeyError, '1' * 40, odb.read, '1' * 40)

    ab = odb.read(BLOB_OID)
    a = odb.read(BLOB_HEX)
    assert ab == a
    assert (GIT_OBJ_BLOB, b'a contents\n') == a

    a2 = odb.read('7f129fd57e31e935c6d60a0c794efe4e6927664b')
    assert (GIT_OBJ_BLOB, b'a contents 2\n') == a2

    a_hex_prefix = BLOB_HEX[:4]
    a3 = odb.read(a_hex_prefix)
    assert (GIT_OBJ_BLOB, b'a contents\n') == a3

def test_write(odb):
    data = b"hello world"
    # invalid object type
    with pytest.raises(ValueError):
        odb.write(GIT_OBJ_ANY, data)

    oid = odb.write(GIT_OBJ_BLOB, data)
    assert type(oid) == Oid



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for reference objects."""

from pathlib import Path

import pytest

from pygit2 import GIT_REF_OID, GIT_REF_SYMBOLIC, Signature
from pygit2 import Commit, Tree, reference_is_valid_name
from pygit2 import AlreadyExistsError, GitError, InvalidSpecError


LAST_COMMIT = '2be5719152d4f82c7302b1c0932d8e5f0a4a0e98'


def test_refs_list_objects(testrepo):
    refs = [(ref.name, ref.target.hex) for ref in testrepo.references.objects]
    assert sorted(refs) == [
        ('refs/heads/i18n', '5470a671a80ac3789f1a6a8cefbcf43ce7af0563'),
        ('refs/heads/master', '2be5719152d4f82c7302b1c0932d8e5f0a4a0e98')]

def test_refs_list(testrepo):
    # Without argument
    assert sorted(testrepo.references) == ['refs/heads/i18n', 'refs/heads/master']

    # We add a symbolic reference
    testrepo.create_reference('refs/tags/version1', 'refs/heads/master')
    assert sorted(testrepo.references) == ['refs/heads/i18n',
                                           'refs/heads/master',
                                           'refs/tags/version1']

def test_head(testrepo):
    head = testrepo.head
    assert LAST_COMMIT == testrepo[head.target].hex
    assert LAST_COMMIT == testrepo[head.raw_target].hex

def test_refs_getitem(testrepo):
    refname = 'refs/foo'
    # Raise KeyError ?
    with pytest.raises(KeyError): testrepo.references[refname]

    # Return None ?
    assert testrepo.references.get(refname) is None

    # Test a lookup
    reference = testrepo.references.get('refs/heads/master')
    assert reference.name == 'refs/heads/master'

def test_refs_get_sha(testrepo):
    reference = testrepo.references['refs/heads/master']
    assert reference.target.hex == LAST_COMMIT

def test_refs_set_sha(testrepo):
    NEW_COMMIT = '5ebeeebb320790caf276b9fc8b24546d63316533'
    reference = testrepo.references.get('refs/heads/master')
    reference.set_target(NEW_COMMIT)
    assert reference.target.hex == NEW_COMMIT

def test_refs_set_sha_prefix(testrepo):
    NEW_COMMIT = '5ebeeebb320790caf276b9fc8b24546d63316533'
    reference = testrepo.references.get('refs/heads/master')
    reference.set_target(NEW_COMMIT[0:6])
    assert reference.target.hex == NEW_COMMIT

def test_refs_get_type(testrepo):
    reference = testrepo.references.get('refs/heads/master')
    assert reference.type == GIT_REF_OID

def test_refs_get_target(testrepo):
    reference = testrepo.references.get('HEAD')
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'

def test_refs_set_target(testrepo):
    reference = testrepo.references.get('HEAD')
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'
    reference.set_target('refs/heads/i18n')
    assert reference.target == 'refs/heads/i18n'
    assert reference.raw_target == b'refs/heads/i18n'

def test_refs_get_shorthand(testrepo):
    reference = testrepo.references.get('refs/heads/master')
    assert reference.shorthand == 'master'
    reference = testrepo.references.create('refs/remotes/origin/master', LAST_COMMIT)
    assert reference.shorthand == 'origin/master'

def test_refs_set_target_with_message(testrepo):
    reference = testrepo.references.get('HEAD')
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'
    sig = Signature('foo', 'bar')
    testrepo.set_ident('foo', 'bar')
    msg = 'Hello log'
    reference.set_target('refs/heads/i18n', message=msg)
    assert reference.target == 'refs/heads/i18n'
    assert reference.raw_target == b'refs/heads/i18n'
    first = list(reference.log())[0]
    assert first.message == msg
    assert first.committer == sig

def test_refs_delete(testrepo):
    # We add a tag as a new reference that points to "origin/master"
    reference = testrepo.references.create('refs/tags/version1', LAST_COMMIT)
    assert 'refs/tags/version1' in testrepo.references

    # And we delete it
    reference.delete()
    assert 'refs/tags/version1' not in testrepo.references

    # Access the deleted reference
    with pytest.raises(GitError): getattr(reference, 'name')
    with pytest.raises(GitError): getattr(reference, 'type')
    with pytest.raises(GitError): getattr(reference, 'target')
    with pytest.raises(GitError): reference.delete()
    with pytest.raises(GitError): reference.resolve()
    with pytest.raises(GitError): reference.rename("refs/tags/version2")

def test_refs_rename(testrepo):
    # We add a tag as a new reference that points to "origin/master"
    reference = testrepo.references.create('refs/tags/version1',
                                            LAST_COMMIT)
    assert reference.name == 'refs/tags/version1'
    reference.rename('refs/tags/version2')
    assert reference.name == 'refs/tags/version2'

#def test_reload(testrepo):
#    name = 'refs/tags/version1'
#    ref = testrepo.create_reference(name, "refs/heads/master", symbolic=True)
#    ref2 = testrepo.lookup_reference(name)
#    ref.delete()
#    assert ref2.name == name
#    with pytest.raises(KeyError): ref2.reload()
#    with pytest.raises(GitError): getattr(ref2, 'name')


def test_refs_resolve(testrepo):
    reference = testrepo.references.get('HEAD')
    assert reference.type == GIT_REF_SYMBOLIC
    reference = reference.resolve()
    assert reference.type == GIT_REF_OID
    assert reference.target.hex == LAST_COMMIT

def test_refs_resolve_identity(testrepo):
    head = testrepo.references.get('HEAD')
    ref = head.resolve()
    assert ref.resolve() is ref

def test_refs_create(testrepo):
    # We add a tag as a new reference that points to "origin/master"
    reference = testrepo.references.create('refs/tags/version1',
                                            LAST_COMMIT)
    refs = testrepo.references
    assert 'refs/tags/version1' in refs
    reference = testrepo.references.get('refs/tags/version1')
    assert reference.target.hex == LAST_COMMIT

    # try to create existing reference
    with pytest.raises(ValueError):
        testrepo.references.create('refs/tags/version1', LAST_COMMIT)

    # try to create existing reference with force
    reference = testrepo.references.create('refs/tags/version1',
                                            LAST_COMMIT, force=True)
    assert reference.target.hex == LAST_COMMIT

def test_refs_create_symbolic(testrepo):
    # We add a tag as a new symbolic reference that always points to
    # "refs/heads/master"
    reference = testrepo.references.create('refs/tags/beta', 'refs/heads/master')
    assert reference.type == GIT_REF_SYMBOLIC
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'

    # try to create existing symbolic reference
    with pytest.raises(ValueError):
        testrepo.references.create('refs/tags/beta', 'refs/heads/master')

    # try to create existing symbolic reference with force
    reference = testrepo.references.create('refs/tags/beta', 'refs/heads/master',
                                           force=True)
    assert reference.type == GIT_REF_SYMBOLIC
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'

#def test_packall_references(testrepo):
#    testrepo.packall_references()


def test_refs_peel(testrepo):
    ref = testrepo.references.get('refs/heads/master')
    assert testrepo[ref.target].id == ref.peel().id
    assert testrepo[ref.raw_target].id == ref.peel().id

    commit = ref.peel(Commit)
    assert commit.tree.id == ref.peel(Tree).id

def test_refs_equality(testrepo):
    ref1 = testrepo.references.get('refs/heads/master')
    ref2 = testrepo.references.get('refs/heads/master')
    ref3 = testrepo.references.get('refs/heads/i18n')

    assert ref1 is not ref2
    assert ref1 == ref2
    assert not ref1 != ref2

    assert ref1 != ref3
    assert not ref1 == ref3

def test_refs_compress(testrepo):
    packed_refs_file = Path(testrepo.path) / 'packed-refs'
    assert not packed_refs_file.exists()
    old_refs = [(ref.name, ref.target.hex)
                for ref in testrepo.references.objects]

    testrepo.references.compress()
    assert packed_refs_file.exists()
    new_refs = [(ref.name, ref.target.hex) for ref in testrepo.references.objects]
    assert old_refs == new_refs


#
# Low level API written in C, repo.references call these.
#

def test_list_all_reference_objects(testrepo):
    repo = testrepo
    refs = [(ref.name, ref.target.hex)
            for ref in repo.listall_reference_objects()]

    assert sorted(refs) == [
        ('refs/heads/i18n', '5470a671a80ac3789f1a6a8cefbcf43ce7af0563'),
        ('refs/heads/master', '2be5719152d4f82c7302b1c0932d8e5f0a4a0e98')]

def test_list_all_references(testrepo):
    repo = testrepo

    # Without argument
    assert sorted(repo.listall_references()) == ['refs/heads/i18n', 'refs/heads/master']
    assert sorted(repo.raw_listall_references()) == [b'refs/heads/i18n', b'refs/heads/master']

    # We add a symbolic reference
    repo.create_reference('refs/tags/version1', 'refs/heads/master')
    assert sorted(repo.listall_references()) == ['refs/heads/i18n',
                                                 'refs/heads/master',
                                                 'refs/tags/version1']
    assert sorted(repo.raw_listall_references()) == [b'refs/heads/i18n',
                                                     b'refs/heads/master',
                                                     b'refs/tags/version1']

def test_lookup_reference(testrepo):
    repo = testrepo

    # Raise KeyError ?
    with pytest.raises(KeyError): repo.lookup_reference('refs/foo')

    # Test a lookup
    reference = repo.lookup_reference('refs/heads/master')
    assert reference.name == 'refs/heads/master'

def test_lookup_reference_dwim(testrepo):
    repo = testrepo

    # remote ref
    reference = testrepo.create_reference('refs/remotes/origin/master', LAST_COMMIT)
    assert reference.shorthand == 'origin/master'
    # tag
    repo.create_reference('refs/tags/version1', LAST_COMMIT)

    # Test dwim lookups

    # Raise KeyError ?
    with pytest.raises(KeyError): repo.lookup_reference_dwim('foo')
    with pytest.raises(KeyError): repo.lookup_reference_dwim('refs/foo')

    reference = repo.lookup_reference_dwim('refs/heads/master')
    assert reference.name == 'refs/heads/master'

    reference = repo.lookup_reference_dwim('master')
    assert reference.name == 'refs/heads/master'

    reference = repo.lookup_reference_dwim('origin/master')
    assert reference.name == 'refs/remotes/origin/master'

    reference = repo.lookup_reference_dwim('version1')
    assert reference.name == 'refs/tags/version1'

def test_resolve_refish(testrepo):
    repo = testrepo

    # remote ref
    reference = testrepo.create_reference('refs/remotes/origin/master', LAST_COMMIT)
    assert reference.shorthand == 'origin/master'
    # tag
    repo.create_reference('refs/tags/version1', LAST_COMMIT)

    # Test dwim lookups

    # Raise KeyError ?
    with pytest.raises(KeyError): repo.resolve_refish('foo')
    with pytest.raises(KeyError): repo.resolve_refish('refs/foo')

    commit, ref = repo.resolve_refish('refs/heads/i18n')
    assert ref.name == 'refs/heads/i18n'
    assert commit.hex == '5470a671a80ac3789f1a6a8cefbcf43ce7af0563'

    commit, ref = repo.resolve_refish('master')
    assert ref.name == 'refs/heads/master'
    assert commit.hex == LAST_COMMIT

    commit, ref = repo.resolve_refish('origin/master')
    assert ref.name == 'refs/remotes/origin/master'
    assert commit.hex == LAST_COMMIT

    commit, ref = repo.resolve_refish('version1')
    assert ref.name == 'refs/tags/version1'
    assert commit.hex == LAST_COMMIT

    commit, ref = repo.resolve_refish(LAST_COMMIT)
    assert ref is None
    assert commit.hex == LAST_COMMIT

    commit, ref = repo.resolve_refish('HEAD~1')
    assert ref is None
    assert commit.hex == '5ebeeebb320790caf276b9fc8b24546d63316533'


def test_reference_get_sha(testrepo):
    reference = testrepo.lookup_reference('refs/heads/master')
    assert reference.target.hex == LAST_COMMIT

def test_reference_set_sha(testrepo):
    NEW_COMMIT = '5ebeeebb320790caf276b9fc8b24546d63316533'
    reference = testrepo.lookup_reference('refs/heads/master')
    reference.set_target(NEW_COMMIT)
    assert reference.target.hex == NEW_COMMIT

def test_reference_set_sha_prefix(testrepo):
    NEW_COMMIT = '5ebeeebb320790caf276b9fc8b24546d63316533'
    reference = testrepo.lookup_reference('refs/heads/master')
    reference.set_target(NEW_COMMIT[0:6])
    assert reference.target.hex == NEW_COMMIT

def test_reference_get_type(testrepo):
    reference = testrepo.lookup_reference('refs/heads/master')
    assert reference.type == GIT_REF_OID

def test_get_target(testrepo):
    reference = testrepo.lookup_reference('HEAD')
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'

def test_set_target(testrepo):
    reference = testrepo.lookup_reference('HEAD')
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'
    reference.set_target('refs/heads/i18n')
    assert reference.target == 'refs/heads/i18n'
    assert reference.raw_target == b'refs/heads/i18n'

def test_get_shorthand(testrepo):
    reference = testrepo.lookup_reference('refs/heads/master')
    assert reference.shorthand == 'master'
    reference = testrepo.create_reference('refs/remotes/origin/master', LAST_COMMIT)
    assert reference.shorthand == 'origin/master'

def test_set_target_with_message(testrepo):
    reference = testrepo.lookup_reference('HEAD')
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'
    sig = Signature('foo', 'bar')
    testrepo.set_ident('foo', 'bar')
    msg = 'Hello log'
    reference.set_target('refs/heads/i18n', message=msg)
    assert reference.target == 'refs/heads/i18n'
    assert reference.raw_target == b'refs/heads/i18n'
    first = list(reference.log())[0]
    assert first.message == msg
    assert first.committer == sig

def test_delete(testrepo):
    repo = testrepo

    # We add a tag as a new reference that points to "origin/master"
    reference = repo.create_reference('refs/tags/version1', LAST_COMMIT)
    assert 'refs/tags/version1' in repo.listall_references()
    assert b'refs/tags/version1' in repo.raw_listall_references()

    # And we delete it
    reference.delete()
    assert 'refs/tags/version1' not in repo.listall_references()
    assert b'refs/tags/version1' not in repo.raw_listall_references()

    # Access the deleted reference
    with pytest.raises(GitError): getattr(reference, 'name')
    with pytest.raises(GitError): getattr(reference, 'type')
    with pytest.raises(GitError): getattr(reference, 'target')
    with pytest.raises(GitError): reference.delete()
    with pytest.raises(GitError): reference.resolve()
    with pytest.raises(GitError): reference.rename("refs/tags/version2")

def test_rename(testrepo):
    # We add a tag as a new reference that points to "origin/master"
    reference = testrepo.create_reference('refs/tags/version1',
                                           LAST_COMMIT)
    assert reference.name == 'refs/tags/version1'
    reference.rename('refs/tags/version2')
    assert reference.name == 'refs/tags/version2'

#   def test_reload(testrepo):
#       name = 'refs/tags/version1'

#       repo = testrepo
#       ref = repo.create_reference(name, "refs/heads/master", symbolic=True)
#       ref2 = repo.lookup_reference(name)
#       ref.delete()
#       assert ref2.name == name
#       with pytest.raises(KeyError): ref2.reload()
#       with pytest.raises(GitError): getattr(ref2, 'name')


def test_reference_resolve(testrepo):
    reference = testrepo.lookup_reference('HEAD')
    assert reference.type == GIT_REF_SYMBOLIC
    reference = reference.resolve()
    assert reference.type == GIT_REF_OID
    assert reference.target.hex == LAST_COMMIT

def test_reference_resolve_identity(testrepo):
    head = testrepo.lookup_reference('HEAD')
    ref = head.resolve()
    assert ref.resolve() is ref

def test_create_reference(testrepo):
    # We add a tag as a new reference that points to "origin/master"
    reference = testrepo.create_reference('refs/tags/version1',
                                           LAST_COMMIT)
    assert 'refs/tags/version1' in testrepo.listall_references()
    assert b'refs/tags/version1' in testrepo.raw_listall_references()
    reference = testrepo.lookup_reference('refs/tags/version1')
    assert reference.target.hex == LAST_COMMIT

    # try to create existing reference
    with pytest.raises(AlreadyExistsError) as error:
        testrepo.create_reference('refs/tags/version1', LAST_COMMIT)
    assert isinstance(error.value, ValueError)

    # Clear error
    del error

    # try to create existing reference with force
    reference = testrepo.create_reference('refs/tags/version1',
                                           LAST_COMMIT, force=True)
    assert reference.target.hex == LAST_COMMIT

def test_create_reference_with_message(testrepo):
    sig = Signature('foo', 'bar')
    testrepo.set_ident('foo', 'bar')
    msg = 'Hello log'
    reference = testrepo.create_reference('refs/heads/feature',
                                          LAST_COMMIT,
                                          message=msg)
    first = list(reference.log())[0]
    assert first.message == msg
    assert first.committer == sig

def test_create_symbolic_reference(testrepo):
    repo = testrepo
    # We add a tag as a new symbolic reference that always points to
    # "refs/heads/master"
    reference = repo.create_reference('refs/tags/beta',
                                      'refs/heads/master')
    assert reference.type == GIT_REF_SYMBOLIC
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'

    # try to create existing symbolic reference
    with pytest.raises(AlreadyExistsError) as error:
        repo.create_reference('refs/tags/beta', 'refs/heads/master')
    assert isinstance(error.value, ValueError)

    # try to create existing symbolic reference with force
    reference = repo.create_reference('refs/tags/beta',
                                      'refs/heads/master', force=True)
    assert reference.type == GIT_REF_SYMBOLIC
    assert reference.target == 'refs/heads/master'
    assert reference.raw_target == b'refs/heads/master'

def test_create_symbolic_reference_with_message(testrepo):
    sig = Signature('foo', 'bar')
    testrepo.set_ident('foo', 'bar')
    msg = 'Hello log'
    reference = testrepo.create_reference('HEAD',
                                          'refs/heads/i18n',
                                          force=True,
                                          message=msg)
    first = list(reference.log())[0]
    assert first.message == msg
    assert first.committer == sig

def test_create_invalid_reference(testrepo):
    repo = testrepo

    # try to create a reference with an invalid name
    with pytest.raises(InvalidSpecError) as error:
        repo.create_reference('refs/tags/in..valid', 'refs/heads/master')
    assert isinstance(error.value, ValueError)

#   def test_packall_references(testrepo):
#       testrepo.packall_references()


def test_peel(testrepo):
    repo = testrepo
    ref = repo.lookup_reference('refs/heads/master')
    assert repo[ref.target].id == ref.peel().id
    assert repo[ref.raw_target].id == ref.peel().id

    commit = ref.peel(Commit)
    assert commit.tree.id == ref.peel(Tree).id



def test_valid_reference_names_ascii():
    assert reference_is_valid_name('HEAD')
    assert reference_is_valid_name('refs/heads/master')
    assert reference_is_valid_name('refs/heads/perfectly/valid')
    assert reference_is_valid_name('refs/tags/v1')
    assert reference_is_valid_name('refs/special/ref')

def test_valid_reference_names_unicode():
    assert reference_is_valid_name('refs/heads/ünicöde')
    assert reference_is_valid_name('refs/tags/😀')

def test_invalid_reference_names():
    assert not reference_is_valid_name('')
    assert not reference_is_valid_name(' refs/heads/master')
    assert not reference_is_valid_name('refs/heads/in..valid')
    assert not reference_is_valid_name('refs/heads/invalid~')
    assert not reference_is_valid_name('refs/heads/invalid^')
    assert not reference_is_valid_name('refs/heads/invalid:')
    assert not reference_is_valid_name('refs/heads/invalid\\')
    assert not reference_is_valid_name('refs/heads/invalid?')
    assert not reference_is_valid_name('refs/heads/invalid[')
    assert not reference_is_valid_name('refs/heads/invalid*')
    assert not reference_is_valid_name('refs/heads/@{no}')
    assert not reference_is_valid_name('refs/heads/foo//bar')

def test_invalid_arguments():
    with pytest.raises(TypeError):
        reference_is_valid_name()
    with pytest.raises(TypeError):
        reference_is_valid_name(None)
    with pytest.raises(TypeError):
        reference_is_valid_name(1)
    with pytest.raises(TypeError):
        reference_is_valid_name('too', 'many')



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

# Standard Library
import functools

# Low level API
from ._pygit2 import *

# High level API
from .blame import Blame, BlameHunk
from .callbacks import git_clone_options, git_fetch_options, get_credentials
from .callbacks import Payload, RemoteCallbacks, CheckoutCallbacks, StashApplyCallbacks
from .config import Config
from .credentials import *
from .errors import check_error, Passthrough
from .ffi import ffi, C
from .index import Index, IndexEntry
from .remote import Remote
from .repository import Repository
from .packbuilder import PackBuilder
from .settings import Settings
from .submodule import Submodule
from .utils import to_bytes, to_str
from ._build import __version__


# Features
features = C.git_libgit2_features()
GIT_FEATURE_THREADS = C.GIT_FEATURE_THREADS
GIT_FEATURE_HTTPS = C.GIT_FEATURE_HTTPS
GIT_FEATURE_SSH = C.GIT_FEATURE_SSH

# GIT_REPOSITORY_INIT_*
GIT_REPOSITORY_INIT_OPTIONS_VERSION = C.GIT_REPOSITORY_INIT_OPTIONS_VERSION
GIT_REPOSITORY_INIT_BARE = C.GIT_REPOSITORY_INIT_BARE
GIT_REPOSITORY_INIT_NO_REINIT = C.GIT_REPOSITORY_INIT_NO_REINIT
GIT_REPOSITORY_INIT_NO_DOTGIT_DIR = C.GIT_REPOSITORY_INIT_NO_DOTGIT_DIR
GIT_REPOSITORY_INIT_MKDIR = C.GIT_REPOSITORY_INIT_MKDIR
GIT_REPOSITORY_INIT_MKPATH = C.GIT_REPOSITORY_INIT_MKPATH
GIT_REPOSITORY_INIT_EXTERNAL_TEMPLATE = C.GIT_REPOSITORY_INIT_EXTERNAL_TEMPLATE
GIT_REPOSITORY_INIT_RELATIVE_GITLINK = C.GIT_REPOSITORY_INIT_RELATIVE_GITLINK
GIT_REPOSITORY_INIT_SHARED_UMASK = C.GIT_REPOSITORY_INIT_SHARED_UMASK
GIT_REPOSITORY_INIT_SHARED_GROUP = C.GIT_REPOSITORY_INIT_SHARED_GROUP
GIT_REPOSITORY_INIT_SHARED_ALL = C.GIT_REPOSITORY_INIT_SHARED_ALL

# GIT_REPOSITORY_OPEN_*
GIT_REPOSITORY_OPEN_NO_SEARCH = C.GIT_REPOSITORY_OPEN_NO_SEARCH
GIT_REPOSITORY_OPEN_CROSS_FS  = C.GIT_REPOSITORY_OPEN_CROSS_FS
GIT_REPOSITORY_OPEN_BARE      = C.GIT_REPOSITORY_OPEN_BARE
GIT_REPOSITORY_OPEN_NO_DOTGIT = C.GIT_REPOSITORY_OPEN_NO_DOTGIT
GIT_REPOSITORY_OPEN_FROM_ENV  = C.GIT_REPOSITORY_OPEN_FROM_ENV

# GIT_ATTR_CHECK_*
GIT_ATTR_CHECK_FILE_THEN_INDEX = C.GIT_ATTR_CHECK_FILE_THEN_INDEX
GIT_ATTR_CHECK_INDEX_THEN_FILE = C.GIT_ATTR_CHECK_INDEX_THEN_FILE
GIT_ATTR_CHECK_INDEX_ONLY      = C.GIT_ATTR_CHECK_INDEX_ONLY
GIT_ATTR_CHECK_NO_SYSTEM       = C.GIT_ATTR_CHECK_NO_SYSTEM

# GIT_FETCH_PRUNE
GIT_FETCH_PRUNE_UNSPECIFIED    = C.GIT_FETCH_PRUNE_UNSPECIFIED
GIT_FETCH_PRUNE                = C.GIT_FETCH_PRUNE
GIT_FETCH_NO_PRUNE             = C.GIT_FETCH_NO_PRUNE

# GIT_CHECKOUT_NOTIFY_*
GIT_CHECKOUT_NOTIFY_NONE       : int = C.GIT_CHECKOUT_NOTIFY_NONE
GIT_CHECKOUT_NOTIFY_CONFLICT   : int = C.GIT_CHECKOUT_NOTIFY_CONFLICT
GIT_CHECKOUT_NOTIFY_DIRTY      : int = C.GIT_CHECKOUT_NOTIFY_DIRTY
GIT_CHECKOUT_NOTIFY_UPDATED    : int = C.GIT_CHECKOUT_NOTIFY_UPDATED
GIT_CHECKOUT_NOTIFY_UNTRACKED  : int = C.GIT_CHECKOUT_NOTIFY_UNTRACKED
GIT_CHECKOUT_NOTIFY_IGNORED    : int = C.GIT_CHECKOUT_NOTIFY_IGNORED
GIT_CHECKOUT_NOTIFY_ALL        : int = C.GIT_CHECKOUT_NOTIFY_ALL

# GIT_STASH_APPLY_PROGRESS_*
GIT_STASH_APPLY_PROGRESS_NONE               : int = C.GIT_STASH_APPLY_PROGRESS_NONE
GIT_STASH_APPLY_PROGRESS_LOADING_STASH      : int = C.GIT_STASH_APPLY_PROGRESS_LOADING_STASH
GIT_STASH_APPLY_PROGRESS_ANALYZE_INDEX      : int = C.GIT_STASH_APPLY_PROGRESS_ANALYZE_INDEX
GIT_STASH_APPLY_PROGRESS_ANALYZE_MODIFIED   : int = C.GIT_STASH_APPLY_PROGRESS_ANALYZE_MODIFIED
GIT_STASH_APPLY_PROGRESS_ANALYZE_UNTRACKED  : int = C.GIT_STASH_APPLY_PROGRESS_ANALYZE_UNTRACKED
GIT_STASH_APPLY_PROGRESS_CHECKOUT_UNTRACKED : int = C.GIT_STASH_APPLY_PROGRESS_CHECKOUT_UNTRACKED
GIT_STASH_APPLY_PROGRESS_CHECKOUT_MODIFIED  : int = C.GIT_STASH_APPLY_PROGRESS_CHECKOUT_MODIFIED
GIT_STASH_APPLY_PROGRESS_DONE               : int = C.GIT_STASH_APPLY_PROGRESS_DONE

# libgit version tuple
LIBGIT2_VER = (LIBGIT2_VER_MAJOR, LIBGIT2_VER_MINOR, LIBGIT2_VER_REVISION)

def init_repository(path, bare=False,
                    flags=GIT_REPOSITORY_INIT_MKPATH,
                    mode=0,
                    workdir_path=None,
                    description=None,
                    template_path=None,
                    initial_head=None,
                    origin_url=None):
    """
    Creates a new Git repository in the given *path*.

    If *bare* is True the repository will be bare, i.e. it will not have a
    working copy.

    The *flags* may be a combination of:

    - GIT_REPOSITORY_INIT_BARE (overriden by the *bare* parameter)
    - GIT_REPOSITORY_INIT_NO_REINIT
    - GIT_REPOSITORY_INIT_NO_DOTGIT_DIR
    - GIT_REPOSITORY_INIT_MKDIR
    - GIT_REPOSITORY_INIT_MKPATH (set by default)
    - GIT_REPOSITORY_INIT_EXTERNAL_TEMPLATE

    The *mode* parameter may be any of GIT_REPOSITORY_SHARED_UMASK (default),
    GIT_REPOSITORY_SHARED_GROUP or GIT_REPOSITORY_INIT_SHARED_ALL, or a custom
    value.

    The *workdir_path*, *description*, *template_path*, *initial_head* and
    *origin_url* are all strings.

    See libgit2's documentation on git_repository_init_ext for further details.
    """
    # Pre-process input parameters
    if path is None:
        raise TypeError('Expected string type for path, found None.')

    if bare:
        flags |= GIT_REPOSITORY_INIT_BARE

    # Options
    options = ffi.new('git_repository_init_options *')
    C.git_repository_init_init_options(options,
                                       GIT_REPOSITORY_INIT_OPTIONS_VERSION)
    options.flags = flags
    options.mode = mode

    if workdir_path:
        workdir_path_ref = ffi.new('char []', to_bytes(workdir_path))
        options.workdir_path = workdir_path_ref

    if description:
        description_ref = ffi.new('char []', to_bytes(description))
        options.description = description_ref

    if template_path:
        template_path_ref = ffi.new('char []', to_bytes(template_path))
        options.template_path = template_path_ref

    if initial_head:
        initial_head_ref = ffi.new('char []', to_bytes(initial_head))
        options.initial_head = initial_head_ref

    if origin_url:
        origin_url_ref = ffi.new('char []', to_bytes(origin_url))
        options.origin_url = origin_url_ref

    # Call
    crepository = ffi.new('git_repository **')
    err = C.git_repository_init_ext(crepository, to_bytes(path), options)
    check_error(err)

    # Ok
    return Repository(to_str(path))


def clone_repository(
        url, path, bare=False, repository=None, remote=None,
        checkout_branch=None, callbacks=None):
    """
    Clones a new Git repository from *url* in the given *path*.

    Returns: a Repository class pointing to the newly cloned repository.

    Parameters:

    url : str
        URL of the repository to clone.
    path : str
        Local path to clone into.
    bare : bool
        Whether the local repository should be bare.
    remote : callable
        Callback for the remote to use.

        The remote callback has `(Repository, name, url) -> Remote` as a
        signature. The Remote it returns will be used instead of the default
        one.
    repository : callable
        Callback for the repository to use.

        The repository callback has `(path, bare) -> Repository` as a
        signature. The Repository it returns will be used instead of creating a
        new one.
    checkout_branch : str
        Branch to checkout after the clone. The default is to use the remote's
        default branch.
    callbacks : RemoteCallbacks
        Object which implements the callbacks as methods.

        The callbacks should be an object which inherits from
        `pyclass:RemoteCallbacks`.
    """

    if callbacks is None:
        callbacks = RemoteCallbacks()

    # Add repository and remote to the payload
    payload = callbacks
    payload.repository = repository
    payload.remote = remote

    with git_clone_options(payload):
        opts = payload.clone_options
        opts.bare = bare

        if checkout_branch:
            checkout_branch_ref = ffi.new('char []', to_bytes(checkout_branch))
            opts.checkout_branch = checkout_branch_ref

        with git_fetch_options(payload, opts=opts.fetch_opts):
            crepo = ffi.new('git_repository **')
            err = C.git_clone(crepo, to_bytes(url), to_bytes(path), opts)
            payload.check_error(err)

    # Ok
    return Repository._from_c(crepo[0], owned=True)


tree_entry_key = functools.cmp_to_key(tree_entry_cmp)

settings = Settings()
